import re
from snappi_ixnetwork.vport import Vport
from collections import namedtuple


class ResourceGroup(object):
    """"""

    """"""

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self._store_properties = []
        self._layer1_conf = None
        self.layer1_check = []
        self._layer1_backup = None

    def set_group(self):
        self.layer1_check = []
        self._store_properties = []
        self._layer1_conf = self._api.snappi_config.get("layer1")
        if (
            self._layer1_conf is None
            or len(self._layer1_conf) == 0
            or self._is_redundant is True
        ):
            return self.layer1_check

        self._cache_properties()
        response = None
        try:
            payload = {"arg1": "/availableHardware", "arg2": []}
            url = (
                "%s/availableHardware/operations/getChassisWithDetailedResouceGroupsInfo"
                % self._api._ixnetwork.href
            )
            response = self._api._request("POST", url, payload)
        except Exception:
            raise Exception(
                "Not able to fetch chassis details. Unable to execute L1 setting"
            )
        self._process_properties(response)

        final_arg2 = []
        error_ports = []
        convert_msgs = []
        idle_msgs = []
        ixn_href = self._api._ixnetwork.href
        for property in self._store_properties:
            # We will change those speed from L1 setting
            if property.aggregate is False:
                continue
            if property.group_mode is None:
                error_ports.append(property.port_name)
                continue
            args = [arg["arg1"] for arg in final_arg2]
            url = property.get_url(ixn_href)
            if url is None:
                idle_msgs.append((property.port_name, property.group_mode))
                continue
            if url not in args:
                convert_msgs.append((property.port_name, property.group_mode))
                arg2 = {"arg1": url, "arg2": property.group_mode}
                final_arg2.append(arg2)

        if len(error_ports) > 0:
            raise Exception(
                "Please check the speed mode or the port numbers for the provided speed mode for these ports ",
                error_ports,
            )
        if len(idle_msgs) > 0:
            self._api.info(
                "Speed conversion is not require for "
                "(port.name, speed) : {0}".format(idle_msgs)
            )
        if len(final_arg2) > 0:
            url = (
                "{0}/availableHardware/operations/"
                "setresourcegroupsinfo".format(ixn_href)
            )
            payload = {
                "arg1": "/availableHardware",
                "arg2": final_arg2,
                "arg3": True,
                "arg4": True,
            }
            try:
                self._api.info(
                    "Setting (port.name, speed) : {0}".format(convert_msgs)
                )
                self._api.info(final_arg2)
                self._api._request("POST", url, payload)
            except:
                # todo: redirect to unknown page. Probable IxNetwork issue
                pass

        return self.layer1_check

    @property
    def _is_redundant(self):
        is_redundant = False
        if self._layer1_backup is not None:
            if self._layer1_conf == self._layer1_backup:
                self._api.info(
                    "Speed change not require "
                    "due to redundant Layer1 config"
                )
                is_redundant = True
                for layer1 in self._layer1_conf:
                    self.layer1_check.append(layer1.get("name"))
        self._layer1_backup = self._layer1_conf
        return is_redundant

    def _cache_properties(self):
        port_list = []
        ports = self._api.snappi_config.get("ports")
        for layer1 in self._layer1_conf:
            port_names = layer1.get("port_names")
            if port_names is None or len(port_names) == 0:
                return
            for port in ports:
                if port in port_list:
                    return
                port_name = port.get("name")
                if port_name in port_names:
                    location = port.get("location")
                    if location is None:
                        raise Exception(
                            "Please configure location to change speed"
                        )
                    location_info = self._api.parse_location_info(location)
                    chassis_info = location_info.chassis_info
                    property = StoreProperty(
                        chassis_info,
                        location_info.card_info,
                        location_info.port_info,
                        port_name,
                        layer1,
                    )
                    self._store_properties.append(property)

    def get_chassis_card(self, property, response):
        ChassisCardInfo = namedtuple(
            "ChassisCardInfo", ["chassis_id", "card_info"]
        )
        chassis_id = 1
        for result in response["result"]:
            chassis_dns = result["dns"]
            card_info = None
            if property.chassis_dns == chassis_dns:
                if int(property.card) == 0:
                    card_info = result["cards"]
                else:
                    for card in result["cards"]:
                        if int(property.card) == int(card["cardId"]):
                            card_info = [card]
                if card_info is not None:
                    return ChassisCardInfo(chassis_id, card_info)
            chassis_id += 1

    def _process_properties(self, response):
        for property in self._store_properties:
            info = self.get_chassis_card(property, response)
            if info is None:
                raise Exception(
                    "Chassis or card not available for %s"
                    % property.get("port_name")
                )
            chassis_id = info.chassis_id
            card_info = info.card_info
            for card in card_info:
                if card["cardAggregationMode"] == "notSupported":
                    continue
                property.aggregate = True
                if self._set_group_mode(property, card, chassis_id) is True:
                    break

    def _set_group_mode(self, property, card, chassis_id):
        for supported_group in card["supportedGroups"]:
            group_id = supported_group["id"]
            current_mode = supported_group["currentSetting"][
                "resourceGroupMode"
            ]
            for available_setting in supported_group["availableSettings"]:
                group_mode = available_setting["resourceGroupMode"]
                for panel_info in available_setting["panelInfo"]:
                    for display_name in panel_info["activePortsDisplayNames"]:
                        if property.port == display_name:
                            l1_name = property.set_property(
                                chassis_id,
                                card,
                                group_id,
                                current_mode,
                                group_mode,
                            )
                            if l1_name is not None:
                                if l1_name not in self.layer1_check:
                                    self.layer1_check.append(l1_name)
                                return True
        return False


class StoreProperty(object):
    def __init__(self, chassis, card, port, port_name, layer1):
        self._chassis = chassis
        self._card = card
        self._port = port.lstrip("0")
        self._port_name = port_name
        self._speed = layer1.speed
        self._l1name = layer1.name
        self._chassis_id = None
        self._card_id = None
        self._group_id = None
        self._current_mode = None
        self._group_mode = None
        self._aggregate = False

    @property
    def chassis_dns(self):
        return self._chassis

    @property
    def card(self):
        return self._card

    @property
    def port(self):
        return self._port

    @property
    def group_mode(self):
        return self._group_mode

    @property
    def port_name(self):
        return self._port_name

    @property
    def l1name(self):
        return self._l1name

    @property
    def aggregate(self):
        return self._aggregate

    @aggregate.setter
    def aggregate(self, value):
        self._aggregate = value

    def _get_speed_map(self, card):
        """This check for specific card"""
        if re.search("maz-xb16", card["appId"].lower()) is not None:
            map_speed = "normal"
        else:
            try:
                map_speed = Vport._SPEED_MODE_MAP[self._speed]
            except Exception:
                raise Exception(
                    "Speed %s not available within internal map" % self._speed
                )
        return map_speed

    def set_property(
        self, chassis_id, card, group_id, current_mode, group_mode
    ):
        map_speed = self._get_speed_map(card)
        if re.search(map_speed, group_mode.lower()) is not None:
            self._chassis_id = chassis_id
            self._card_id = card["cardId"]
            self._group_id = group_id
            self._current_mode = current_mode
            self._group_mode = group_mode
            return self._l1name
        return None

    def get_url(self, ixn_href):
        if self._current_mode == self._group_mode:
            return None
        url = (
            "{0}/availableHardware/chassis/{1}"
            "/card/{2}/aggregation/{3}".format(
                ixn_href, self._chassis_id, self._card_id, self._group_id
            )
        )
        return url



from snappi_ixnetwork.device.base import Base
from snappi_ixnetwork.logger import get_ixnet_logger


class Ethernet(Base):
    _ETHERNET = {
        "mac": "mac",
        "mtu": "mtu",
    }

    _VLAN = {
        "tpid": {
            "ixn_attr": "tpid",
            "enum_map": {
                "x8100": "ethertype8100",
                "x88a8": "ethertype88a8",
                "x9100": "ethertype9100",
                "x9200": "ethertype9200",
                "x9300": "ethertype9300",
            },
        },
        "priority": "priority",
        "id": "vlanId",
    }

    _IP = {"address": "address", "gateway": "gatewayIp", "prefix": "prefix"}

    _GATEWAY_MAC = {
        "gateway_mac": "manualGatewayMac",
    }

    def __init__(self, ngpf):
        super(Ethernet, self).__init__()
        self._ngpf = ngpf
        self.logger = get_ixnet_logger(__name__)

    def config(self, ethernet, ixn_dg):
        ixn_eth = self.create_node_elemet(
            ixn_dg, "ethernet", ethernet.get("name")
        )
        self._ngpf.set_device_info(ethernet, ixn_eth)
        self.configure_multivalues(ethernet, ixn_eth, Ethernet._ETHERNET)
        vlans = ethernet.get("vlans")
        if vlans is not None and len(vlans) > 0:
            ixn_eth["enableVlans"] = self.multivalue(True)
            ixn_eth["vlanCount"] = len(vlans)
            self._configure_vlan(ixn_eth, vlans)
        self._configure_ipv4(ixn_eth, ethernet)
        self._configure_ipv6(ixn_eth, ethernet)

    def _configure_vlan(self, ixn_eth, vlans):
        self.logger.debug("Configuring VLAN")
        for vlan in vlans:
            ixn_vlan = self.create_node_elemet(
                ixn_eth, "vlan", vlan.get("name")
            )
            self.configure_multivalues(vlan, ixn_vlan, Ethernet._VLAN)

    def _configure_ipv4(self, ixn_eth, ethernet):
        self.logger.debug("Configuring IPv4 interface")
        ipv4_addresses = ethernet.get("ipv4_addresses")
        if ipv4_addresses is None:
            return

        eth_name = ethernet.name
        if eth_name not in self._ngpf.ether_v4gateway_map:
            self._ngpf.ether_v4gateway_map[eth_name] = []

        for ipv4_address in ipv4_addresses:
            self._ngpf.ether_v4gateway_map[eth_name].append(
                ipv4_address.gateway
            )
            if self._ngpf.is_ip_restricted(eth_name) == True:
                ixn_ip = None
                self._ngpf.set_device_info(ipv4_address, ixn_ip)
            else:
                ixn_ip = self.create_node_elemet(
                    ixn_eth, "ipv4", ipv4_address.get("name")
                )
                self._ngpf.set_device_info(ipv4_address, ixn_ip)
                self.configure_multivalues(ipv4_address, ixn_ip, Ethernet._IP)
                if ipv4_address.gateway_mac.choice == "value":
                    self.configure_multivalues_with_choice(
                        ipv4_address, ixn_ip, Ethernet._GATEWAY_MAC
                    )

    def _configure_ipv6(self, ixn_eth, ethernet):
        self.logger.debug("Configuring IPv6 interface")
        ipv6_addresses = ethernet.get("ipv6_addresses")
        if ipv6_addresses is None:
            return

        eth_name = ethernet.name
        if eth_name not in self._ngpf.ether_v6gateway_map:
            self._ngpf.ether_v6gateway_map[eth_name] = []

        for ipv6_address in ipv6_addresses:
            self._ngpf.ether_v6gateway_map[eth_name].append(
                ipv6_address.gateway
            )
            if self._ngpf.is_ip_restricted(eth_name) == True:
                ixn_ip = None
                self._ngpf.set_device_info(ipv6_address, ixn_ip)
            else:
                ixn_ip = self.create_node_elemet(
                    ixn_eth, "ipv6", ipv6_address.get("name")
                )
                self._ngpf.set_device_info(ipv6_address, ixn_ip)
                self.configure_multivalues(ipv6_address, ixn_ip, Ethernet._IP)
                if ipv6_address.gateway_mac.choice == "value":
                    self.configure_multivalues_with_choice(
                        ipv6_address, ixn_ip, Ethernet._GATEWAY_MAC
                    )



from snappi_ixnetwork.timer import Timer


class Ping(object):
    """Transforms OpenAPI objects into IxNetwork objects

    Args
    ----
    - ixnetworkapi (Api): instance of the Api class

    Process
    -------
    - Ping to the interfaces and return response

    Notes
    -----
    """

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi

    def results(self, ping_request, req_type=None):
        responses = []
        v4_names = []
        for device in self._api._config.devices:
            for eth in device.ethernets:
                for ip in eth.ipv4_addresses:
                    v4_names.append(ip.name)
        v6_names = []
        for device in self._api._config.devices:
            for eth in device.ethernets:
                for ip in eth.ipv6_addresses:
                    v6_names.append(ip.name)
        if req_type == None:
            raise Exception(
                "req_type variable is mandatory to decide ipv4 or ipv6"
            )
        else:
            with Timer(self._api, "Ping requests completed in"):
                for endpoint in ping_request.requests:
                    response = {}
                    src_name = endpoint.get("src_name")
                    dst_ip = endpoint.get("dst_ip")
                    if req_type == "ipv4":
                        if src_name not in v4_names:
                            msg = (
                                src_name
                                + """ is not available in the configured v4 interface names """
                                + str(v4_names)
                            )
                            raise Exception(msg)
                        ip_obj = (
                            self._api._ixnetwork.Topology.find()
                            .DeviceGroup.find()
                            .Ethernet.find()
                            .Ipv4.find(Name=src_name)
                        )
                    elif req_type == "ipv6":
                        if src_name not in v6_names:
                            msg = (
                                src_name
                                + """ is not available in the configured v6 interface names """
                                + str(v6_names)
                            )
                            raise Exception(msg)
                        ip_obj = (
                            self._api._ixnetwork.Topology.find()
                            .DeviceGroup.find()
                            .Ethernet.find()
                            .Ipv6.find(Name=src_name)
                        )
                    self._api.info("Sending ping to %s" % dst_ip)
                    ping_status = ip_obj.SendPing(DestIP=dst_ip)
                    for reply in ping_status:
                        if dst_ip in reply["arg3"]:
                            if reply["arg2"]:
                                response["result"] = "succeeded"
                            else:
                                response["result"] = "failed"
                    response["src_name"] = src_name
                    response["dst_ip"] = dst_ip
                    responses.append(response)
                return responses



from snappi_ixnetwork.device.base import Base
from snappi_ixnetwork.device.mka import Mka
from snappi_ixnetwork.logger import get_ixnet_logger


class Macsec(Base):
    _CIPHER = {
        "cipher_suite": {
            "ixn_attr": "cipherSuite",
            "enum_map": {
                "gcm_aes_128": "aes128",
                "gcm_aes_256": "aes256",
                "gcm_aes_xpn_128": "aesxpn128",
                "gcm_aes_xpn_256": "aesxpn256",
            },
        },
        "confidentiality": "enableConfidentiality",
        "confidentiality_offset": {
            "ixn_attr": "confidentialityOffset",
            "enum_map": {"zero": 0, "thirty": 30, "fifty": 50},
        },
    }

    _DATA_PLANE_TX = {
        "end_station": "endStation",
        "include_sci": "includeSci",
    }

    _STATIC_KEY_TXSC = {
        "system_id": "systemId",
        "port_id": "portId",
    }

    _STATIC_KEY_RXSC = {
        "dut_sci_system_id": "dutSciMac",
        "dut_sci_port_id": "dutSciPortId",
        "dut_msb_xpn": "dutMsbOfXpn",
    }

    def __init__(self, ngpf):
        super(Macsec, self).__init__()
        self._ngpf = ngpf
        self._mka = Mka(ngpf)
        self.is_dynamic_key = False
        self.logger = get_ixnet_logger(__name__)

    def config(self, device):
        self.logger.debug("Configuring MACsec")
        macsec = device.get("macsec")
        if macsec is None:
            return
        self.is_dynamic_key = self._is_dynamic_key(macsec)
        if self.is_dynamic_key == True:
            self._mka.config(device)
        self.is_data_plane = self._is_data_plane(macsec)
        self._config_ethernet_interfaces(device)

    def _is_valid(self, ethernet_name, secy):
        is_valid = True
        if is_valid == True:
            self.logger.debug("MACsec validation success")
        else:
            self.logger.debug("MACsec validation failure")
        return is_valid

    def set_ip_restriction(self, device):
        self.logger.debug("Checking if IPv4/ v6 is restricted")
        macsec = device.get("macsec")
        if macsec is None:
            return
        ethernet_interfaces = macsec.get("ethernet_interfaces")
        if ethernet_interfaces is None:
            return
        for ethernet_interface in ethernet_interfaces:
            secy = ethernet_interface.get("secure_entity")
            if secy is None:
                continue
            data_plane = secy.get("data_plane")
            if data_plane is None:
                continue
            encapsulation = data_plane.get("encapsulation")
            if encapsulation is None:
                continue
            crypto_engine = encapsulation.get("crypto_engine")
            if crypto_engine is None:
                continue
            if crypto_engine.choice == "encrypt_only":
                self.logger.debug("IPv4/ v6 is restricted on ethernet %s as encrypt_only MACsec engine is configured" % ethernet_interface.get("eth_name"))
                self._ngpf.set_ip_restricted(ethernet_interface.get("eth_name"))

    def _is_dynamic_key(self, macsec):
        self.logger.debug("Checking if MKA is configured")
        is_mka = False
        ethernet_interfaces = macsec.get("ethernet_interfaces")
        for ethernet_interface in ethernet_interfaces:
            secy = ethernet_interface.get("secure_entity")
            key_generation_protocol = secy.get("key_generation_protocol")
            protocol = key_generation_protocol.choice
            if protocol == "mka":
                is_mka = True
                break
        if is_mka == True:
            self.logger.info("MKA is configured in test")
        return is_mka

    def _is_data_plane(self, macsec):
        self.logger.debug("Checking if MACsec data plane is configured")
        is_data_plane = False
        ethernet_interfaces = macsec.get("ethernet_interfaces")
        for ethernet_interface in ethernet_interfaces:
            ethernet_interface = ethernet_interfaces[0]
            secy = ethernet_interface.get("secure_entity")
            data_plane = secy.get("data_plane")
            if data_plane is None:
                continue
            if data_plane.choice == "encapsulation":
                is_data_plane = True
                break
        return is_data_plane

    def _config_ethernet_interfaces(self, device):
        self.logger.debug("Configuring MACsec interfaces")
        macsec = device.get("macsec")
        ethernet_interfaces = macsec.get("ethernet_interfaces")
        if ethernet_interfaces is None:
            return
        for ethernet_interface in ethernet_interfaces:
            ethernet_name = ethernet_interface.get("eth_name")
            self._ngpf.working_dg = self._ngpf.api.ixn_objects.get_working_dg(
                ethernet_name
            )
            if not self._is_valid(
                ethernet_name, ethernet_interface.get("secure_entity")
            ):
                continue
            self._config_secy(device, ethernet_interface)

    def _config_secy(self, device, ethernet_interface):
        self.logger.debug("Configuring SecY")
        secy = ethernet_interface.get("secure_entity")
        if secy is None:
            return
        ethernet_name = ethernet_interface.get("eth_name")
        ixn_ethernet = self._ngpf.api.ixn_objects.get_object(ethernet_name)
        # if self.is_dynamic_key:
        # config MKA
        if self.is_data_plane:
            if (
                secy.data_plane.encapsulation.crypto_engine.choice
                == "encrypt_only"
            ):
                self.logger.debug("Configuring SecY: encrypt_only")
                ixn_staticmacsec = self.create_node_elemet(
                    ixn_ethernet, "staticMacsec", secy.get("name")
                )
                self._ngpf.set_device_info(secy, ixn_staticmacsec)
                if not self.is_dynamic_key:
                    self._config_static_key(secy, ixn_staticmacsec)
                self._config_data_plane(
                    device, ethernet_interface, ixn_staticmacsec
                )

    def _config_static_key(self, secy, ixn_staticmacsec):
        self.logger.debug("Configuring static key")
        key_generation_protocol = secy.get("key_generation_protocol")
        static_key = key_generation_protocol.get("static_key")
        if static_key is None:
            return
        self._config_static_key_cipher(static_key, ixn_staticmacsec)
        self._config_static_key_tx(static_key, ixn_staticmacsec)
        self._config_static_key_rx(static_key, ixn_staticmacsec)

    def _config_static_key_cipher(self, static_key, ixn_staticmacsec):
        self.logger.debug("Configuring cipher for static key")
        self.configure_multivalues(
            static_key, ixn_staticmacsec, Macsec._CIPHER
        )

    def _config_data_plane(self, device, ethernet_interface, ixn_staticmacsec):
        self.logger.debug("Configuring data plane properties")
        secy = ethernet_interface.get("secure_entity")
        data_plane = secy.get("data_plane")
        encapsulation = data_plane.get("encapsulation")
        if encapsulation is None:
            return
        self._config_data_plane_tx(encapsulation, ixn_staticmacsec)
        self._config_data_plane_rx(encapsulation, ixn_staticmacsec)
        self._config_crypto_engine(
            device, ethernet_interface, ixn_staticmacsec
        )

    def _config_data_plane_tx(self, encapsulation, ixn_staticmacsec):
        self.logger.debug("Configuring data plane Tx properties")
        tx = encapsulation.get("tx")
        if tx is None:
            return
        self.logger.debug(
            "end_station %s include_sci %s" % (tx.end_station, tx.include_sci)
        )
        self.configure_multivalues(tx, ixn_staticmacsec, Macsec._DATA_PLANE_TX)

    def _config_data_plane_rx(self, encapsulation, ixn_staticmacsec):
        self.logger.debug("Configuring data plane Rx properties")
        rx = encapsulation.get("rx")
        if rx is None:
            return
        # TODO: replay protection, replay window.
        # self.logger.debug("replay_protection %s replay_window %s" % (rx.replay_protection, rx.replay_window))

    def _config_crypto_engine(
        self, device, ethernet_interface, ixn_staticmacsec
    ):
        self.logger.debug("Configuring crypto engine properties")
        secy = ethernet_interface.get("secure_entity")
        data_plane = secy.get("data_plane")
        encapsulation = data_plane.get("encapsulation")
        crypto_engine = encapsulation.get("crypto_engine")
        if crypto_engine is None:
            return
        if crypto_engine.choice == "encrypt_only":
            self._config_crypto_engine_encrypt_only(
                device, ethernet_interface, ixn_staticmacsec
            )

    def _config_static_key_tx(self, static_key, ixn_staticmacsec):
        self.logger.debug("Configuring Tx properties for static key")
        tx = static_key.get("tx")
        if tx is None:
            return
        self._config_static_key_txsc(static_key, ixn_staticmacsec)
        self._config_rekey_mode(static_key.tx.rekey_mode, ixn_staticmacsec)

    def _config_static_key_rx(self, static_key, ixn_staticmacsec):
        self.logger.debug("Configuring Rx properties for static key")
        rx = static_key.get("rx")
        if rx is None:
            return
        self._config_static_key_rxsc(static_key, ixn_staticmacsec)

    def _config_static_key_txsc(self, static_key, ixn_staticmacsec):
        self.logger.debug("Configuring Tx secure channels for static key")
        tx = static_key.get("tx")
        txscs = tx.get("secure_channels")
        txsc = txscs[0]
        self.configure_multivalues(
            txsc, ixn_staticmacsec, Macsec._STATIC_KEY_TXSC
        )
        tx_saks = txsc.saks
        ixn_staticmacsec["txSakPoolSize"] = len(tx_saks)
        saks = []
        sscis = []
        salts = []
        for tx_sak in tx_saks:
            saks.append(tx_sak.sak)
            sscis.append(tx_sak.ssci)
            salts.append(tx_sak.salt)
        ixn_tx_sak_pool = self.create_node_elemet(
            ixn_staticmacsec, "txSakPool", name=None
        )
        cipher_suite = static_key.cipher_suite
        if cipher_suite == "gcm_aes_128" or cipher_suite == "gcm_aes_xpn_128":
            ixn_tx_sak_pool["txSak128"] = self.multivalue(saks)
        elif (
            cipher_suite == "gcm_aes_256" or cipher_suite == "gcm_aes_xpn_256"
        ):
            ixn_tx_sak_pool["txSak256"] = self.multivalue(saks)
        if (
            cipher_suite == "gcm_aes_xpn_128"
            or cipher_suite == "gcm_aes_xpn_256"
        ):
            ixn_tx_sak_pool["txSsci"] = self.multivalue(sscis)
            ixn_tx_sak_pool["txSalt"] = self.multivalue(salts)

    def _config_static_key_rxsc(self, static_key, ixn_staticmacsec):
        self.logger.debug("Configuring Rx secure channels for static key")
        rx = static_key.get("rx")
        rxscs = rx.get("secure_channels")
        rxsc = rxscs[0]
        self.configure_multivalues(
            rxsc, ixn_staticmacsec, Macsec._STATIC_KEY_RXSC
        )
        rx_saks = rxsc.saks
        ixn_staticmacsec["rxSakPoolSize"] = len(rx_saks)
        saks = []
        sscis = []
        salts = []
        for rx_sak in rx_saks:
            saks.append(rx_sak.sak)
            sscis.append(rx_sak.ssci)
            salts.append(rx_sak.salt)
        ixn_rx_sak_pool = self.create_node_elemet(
            ixn_staticmacsec, "rxSakPool", name=None
        )
        static_key = secy.get("static_key")
        cipher_suite = static_key.cipher_suite
        if cipher_suite == "gcm_aes_128" or cipher_suite == "gcm_aes_xpn_128":
            ixn_rx_sak_pool["rxSak128"] = self.multivalue(saks)
        elif (
            cipher_suite == "gcm_aes_256" or cipher_suite == "gcm_aes_xpn_256"
        ):
            ixn_rx_sak_pool["rxSak256"] = self.multivalue(saks)
        if (
            cipher_suite == "gcm_aes_xpn_128"
            or cipher_suite == "gcm_aes_xpn_256"
        ):
            ixn_rx_sak_pool["rxSsci"] = self.multivalue(sscis)
            ixn_rx_sak_pool["rxSalt"] = self.multivalue(salts)

    def _config_rekey_mode(self, rekey_mode, ixn_staticmacsec):
        self.logger.debug("Configuring Tx rekey for static key")
        if rekey_mode.choice == "dont_rekey":
            # ixn_staticmacsec["rekeyMode"] = "timerBased"
            ixn_staticmacsec["rekeyBehaviour"] = "dontRekey"
        elif rekey_mode.choice == "timer_based":
            timer_based = rekey_mode.timer_based
            # ixn_staticmacsec["rekeyMode"] = "timerBased"
            if timer_based.choice == "fixed_count":
                ixn_staticmacsec["rekeyBehaviour"] = "rekeyFixedCount"
                ixn_staticmacsec["periodicRekeyAttempts"] = (
                    timer_based.fixed_count
                )
                ixn_staticmacsec["periodicRekeyInterval"] = (
                    timer_based.interval
                )
            elif timer_based.choice == "continuous":
                ixn_staticmacsec["rekeyBehaviour"] = "rekeyContinuous"
                ixn_staticmacsec["periodicRekeyInterval"] = (
                    timer_based.interval
                )
        # elif rekey_mode.choice == "pn_based":
        #    ixn_staticmacsec["rekeyMode"] = "pnBased"
        #    ixn_staticmacsec["rekeyBehaviour"] = "rekeyContinuous"

    def _config_crypto_engine_encrypt_only(
        self, device, ethernet_interface, ixn_staticmacsec
    ):
        secy = ethernet_interface.get("secure_entity")
        self._config_crypto_engine_encrypt_only_tx_pn(secy, ixn_staticmacsec)
        self._config_crypto_engine_encrypt_only_traffic(
            device, ethernet_interface, ixn_staticmacsec
        )

    def _config_crypto_engine_encrypt_only_traffic(
        self, device, ethernet_interface, ixn_staticmacsec
    ):
        ethernet_name = ethernet_interface.get("eth_name")
        secy = ethernet_interface.get("secure_entity")
        self.logger.debug(
            "Configuring stateless encryption traffic from ethernet %s secy %s"
            % (ethernet_name, secy.get("name"))
        )
        ethernets = device.get("ethernets")
        for ethernet in ethernets:
            if ethernet.get("name") == ethernet_name:
                self._ngpf.api.set_device_traffic_endpoint(
                    ethernet_name, secy.get("name")
                )
                ipv4_addresses = ethernet.get("ipv4_addresses")
                if ipv4_addresses is None:
                    # raise Exception("IPv4 not configured on ethernet %s" % ethernet_name)
                    self.logger.info(
                        "IPv4 not configured on ethernet %s" % ethernet_name
                    )
                    break
                elif len(ipv4_addresses) > 1:
                    raise Exception(
                        "More than one IPv4 address configured on ethernet %s"
                        % ethernet_name
                    )
                ipv4_address = ipv4_addresses[0].address
                ipv4_gateway_mac = ipv4_addresses[0].gateway_mac
                if ipv4_gateway_mac is None:
                    raise Exception(
                        "IPv4 gateway mac not configured for IPv4 address %s on ethernet %s"
                        % (ethernet_name, ipv4_address)
                    )
                elif (
                    not ipv4_gateway_mac.choice == "value"
                    or ipv4_gateway_mac.value is None
                ):
                    raise Exception(
                        "IPv4 gateway static mac not configured for IPv4 address %s on ethernet %s"
                        % (ethernet_name, ipv4_address)
                    )
                ipv4_gateway_static_mac = ipv4_addresses[0].gateway_mac.value
                ixn_staticmacsec["sourceIp"] = self.multivalue(ipv4_address)
                ixn_staticmacsec["dutMac"] = self.multivalue(
                    ipv4_gateway_static_mac
                )
                self._ngpf.api.set_device_traffic_endpoint(
                    ipv4_addresses[0].name, secy.get("name")
                )
                break

    def _config_crypto_engine_encrypt_only_tx_pn(self, secy, ixn_staticmacsec):
        crypto_engine = secy.data_plane.encapsulation.crypto_engine
        crypto_engine_encrypt_only = crypto_engine.encrypt_only
        if crypto_engine_encrypt_only is None:
            return
        txscs = crypto_engine_encrypt_only.secure_channels
        if len(txscs) > 0:
            txsc = txscs[0]
        else:
            return
        if txsc is None:
            return
        engine_tx_pn = txsc.get("tx_pn")
        if engine_tx_pn is None:
            return
        self.logger.debug(
            "Configuring Tx PN of stateless encryption only engine secy %s"
            % secy.get("name")
        )
        if engine_tx_pn.choice == "fixed_pn":
            ixn_staticmacsec["incrementingPn"] = False
            engine_tx_pn_fixed = engine_tx_pn.fixed
            ixn_staticmacsec["fixedPn"] = engine_tx_pn_fixed.pn
            ixn_staticmacsec["mvFixedXpn"] = self.multivalue(
                engine_tx_pn_fixed.xpn
            )
        elif engine_tx_pn.choice == "incrementing_pn":
            ixn_staticmacsec["incrementingPn"] = True
            engine_tx_pn_incr = engine_tx_pn.incrementing
            ixn_staticmacsec["packetCountPn"] = engine_tx_pn_incr.count
            ixn_staticmacsec["firstPn"] = engine_tx_pn_incr.starting_pn
            ixn_staticmacsec["mvFirstXpn"] = self.multivalue(
                engine_tx_pn_incr.starting_xpn
            )



import json
import re
import time
import logging
from collections import namedtuple

import snappi
from ixnetwork_restpy import TestPlatform, SessionAssistant

from snappi_ixnetwork.logger import setup_ixnet_logger
from snappi_ixnetwork.capture import Capture
from snappi_ixnetwork.device.ngpf import Ngpf
from snappi_ixnetwork.exceptions import SnappiIxnException
from snappi_ixnetwork.lag import Lag
from snappi_ixnetwork.objectdb import IxNetObjects
from snappi_ixnetwork.ping import Ping
from snappi_ixnetwork.protocolmetrics import ProtocolMetrics
from snappi_ixnetwork.resourcegroup import ResourceGroup
from snappi_ixnetwork.timer import Timer
from snappi_ixnetwork.trafficitem import TrafficItem
from snappi_ixnetwork.validation import Validation
from snappi_ixnetwork.vport import Vport
from snappi_ixnetwork.ixnetworkconfig import IxNetworkConfig
from snappi_ixnetwork.device.mka import Mka
from snappi_ixnetwork.device.macsec import Macsec


class Api(snappi.Api):
    """IxNetwork implementation of the abstract-open-traffic-generator package

    Args
    ----
    - address (str): The address of the IxNetwork API Server
    - port (str): The rest port of the IxNetwork API Server
    - username (str): The username for Linux IxNetwork API Server
        This is not required when connecting to single session environments
    - password (str): The password for Linux IxNetwork API Server
        This is not required when connecting to single session environments
    """

    _CONVERGENCE = {
        ("data_plane_convergence_us", "DP/DP Convergence Time (us)", float),
        (
            "control_plane_data_plane_convergence_us",
            "CP/DP Convergence Time (us)",
            float,
        ),
    }

    _DP_CONVERGENCE = {
        ("data_plane_convergence_us", "DP/DP Convergence Time (us)", float),
    }

    _EVENT = {
        ("begin_timestamp_ns", "Event Start Timestamp", int),
        ("end_timestamp_ns", "Event End Timestamp", int),
    }

    _TRIGGERED_EVENT = ""

    def __init__(self, **kwargs):
        """Create a session
        - address (str): The ip address of the TestPlatform to connect to
        where test sessions will be created or connected to.
        - port (str): The rest port of the TestPlatform to connect to.
        - username (str): The username to be used for authentication
        - password (str): The password to be used for authentication
        """
        super(Api, self).__init__()
        location = kwargs.get("location")
        username = kwargs.get("username")
        password = kwargs.get("password")
        license_servers = kwargs.get("license_servers")
        self._log_level = (
            logging.INFO
            if kwargs.get("loglevel") is None
            else kwargs.get("loglevel")
        )
        self.logger = setup_ixnet_logger(self.log_level, module_name=__name__)
        location = "https://127.0.0.1:11009" if location is None else location
        self._address, self._port = self._get_addr_port(location)
        self._username = "admin" if username is None else username
        self._password = "admin" if password is None else password
        self._license_servers = (
            [] if license_servers is None else license_servers
        )
        self._running_config = None
        self._config = None
        self._assistant = None
        self._ixn_errors = list()
        self._config_objects = {}
        self._device_traffic_endpoint = {}
        self._device_encap = {}
        self.ixn_objects = None
        self._config_type = self.config()
        self._control_state = self.control_state()
        self._control_action = self.control_action()
        self._flows_update = self.config_update()
        self._flows_delete = self.config_delete()
        self._flows_append = self.config_append()
        self._capture_request = self.capture_request()
        self.ixn_routes = []
        self.validation = Validation(self)
        self.vport = Vport(self)
        self.lag = Lag(self)
        self.ngpf = Ngpf(self)
        self.traffic_item = TrafficItem(self)
        self.capture = Capture(self)
        self.ping = Ping(self)
        self.protocol_metrics = ProtocolMetrics(self)
        self.resource_group = ResourceGroup(self)
        self.do_compact = False
        self._dev_compacted = {}
        self._previous_errors = []
        self._initial_flows_config = None
        self._flow_tracking = False
        self._convergence_timeout = 3
        self._event_info = None
        self._ixnet_specific_config = None
        self._mka = Mka(self)
        self._macsec = Macsec(self)

        self._ixn_route_info = namedtuple(
            "IxnRouteInfo", ["ixn_obj", "index", "multiplier"]
        )

    def _get_addr_port(self, host):
        items = host.split("/")
        items = items[-1].split(":")

        addr = items[0]
        if len(items) == 2:
            return addr, items[-1]
        else:
            if host.startswith("https"):
                return addr, "443"
            else:
                return addr, "80"

    @property
    def log_level(self):
        return self._log_level

    def enable_scaling(self, do_compact=False):
        self.do_compact = do_compact

    def _enable_flow_tracking(self, _flow_tracking=False):
        self._flow_tracking = _flow_tracking

    @property
    def snappi_config(self):
        return self._config

    def get_config_object(self, name):
        try:
            return self._config_objects[name]
        except KeyError:
            raise NameError("snappi object named {0} not found".format(name))

    def get_device_encap(self, name):
        try:
            return self._device_encap[name]
        except KeyError:
            raise NameError("snappi object named {0} not found".format(name))

    def set_device_encap(self, name, type):
        self._device_encap[name] = type

    def get_device_traffic_endpoint(self, name):
        try:
            value = self._device_traffic_endpoint[name]
        except KeyError:
            value = None
        return value

    def set_device_traffic_endpoint(self, name, value):
        self._device_traffic_endpoint[name] = value

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        self._username = value

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password = value

    @property
    def assistant(self):
        return self._assistant

    @property
    def dev_compacted(self):
        return self._dev_compacted

    @property
    def ixnet_specific_config(self):
        if self._ixnet_specific_config is None:
            self._ixnet_specific_config = IxNetworkConfig(self)

        return self._ixnet_specific_config

    def set_dev_compacted(self, dev_name, name_list):
        for index, name in enumerate(name_list):
            self._dev_compacted[name] = {"dev_name": dev_name, "index": index}

    def _dict_to_obj(self, source):
        """Returns an object given a dict"""
        if isinstance(source, list):
            source = [self._dict_to_obj(x) for x in source]
        if not isinstance(source, dict):
            return source
        o = lambda: None
        for k, v in source.items():
            o.__dict__[k] = self._dict_to_obj(v)
        return o

    def _request_detail(self):
        request_detail = snappi.Warning()
        errors = self._errors
        warnings = list()
        app_errors = self._globals.AppErrors.find()
        if len(app_errors) > 0:
            current_errors = app_errors[0].Error.find()
            if len(current_errors) > 0:
                for error in current_errors:
                    # Add loop as sequence are not sorted
                    match = [
                        o
                        for o in self._ixn_errors
                        if o.Name == error.Name
                        and o.LastModified == error.LastModified
                    ]
                    if len(match) == 0:
                        if error.ErrorLevel == "kWarning":
                            warning = "IxNet - {0}".format(
                                error.Name
                                if error.Description == ""
                                else error.Description
                            )
                            self.info(warning)
                            warnings.append(warning)
                        if error.ErrorLevel == "kError":
                            error = "IxNet - {0}".format(
                                error.Name
                                if error.Description == ""
                                else error.Description
                            )
                            self.info(error)
                            errors.append(error)

        # Need to add this hack as dumping json config throws this exception
        # Traffic regeneration will be part of set_transmit_state
        errors = [
            error
            for error in errors
            if "before trying to generate BGP EVPN traffic" not in error
        ]

        if len(errors) > 0:
            self._errors = []
            raise SnappiIxnException(500, errors)
        request_detail.warnings = warnings
        return request_detail

    def set_config(self, config):
        """Set, update, append or delete the configuration"""
        try:
            if isinstance(config, (type(self._config_type), str)) is False:
                raise TypeError(
                    "The content must be of type Union[Config, str]"
                )

            if isinstance(config, str) is True:
                config = self._config_type.deserialize(config)
            self.config_ixnetwork(config)
            # CP-DP Convergence config
            ixn_cpdpconvergence = self._traffic.Statistics.CpdpConvergence
            ixn_cpdpconvergence.Enabled = False
            cfg = config.get("events")
            if cfg is not None:
                cp_events = cfg.get("cp_events")
                if cp_events is not None:
                    cp_events_enable = cp_events.get("enable")
                else:
                    cp_events_enable = False
                dp_events = cfg.get("dp_events")
                if dp_events is not None:
                    dp_events_enable = dp_events.get("enable")
                    rx_rate_threshold = dp_events.get("rx_rate_threshold")
                else:
                    dp_events_enable = False
                # Enable cp-dp convergence if any one of cp or dp is true
                if cp_events_enable or dp_events_enable:
                    ixn_cpdpconvergence.Enabled = True
                    # For CP events
                    if cp_events_enable:
                        ixn_cpdpconvergence.EnableControlPlaneEvents = True
                    # For DP events
                    if dp_events_enable:
                        if self.traffic_item.has_latency is True:
                            raise Exception(
                                "We are supporting either latency or dp convergence"
                            )
                        ixn_cpdpconvergence.EnableDataPlaneEventsRateMonitor = (
                            True
                        )
                        ixn_cpdpconvergence.DataPlaneThreshold = (
                            rx_rate_threshold
                        )

                for ixn_traffic_item in self._traffic_item.find():
                    ixn_traffic_item.Tracking.find()[0].TrackBy = [
                        "destEndpoint0",
                        "destSessionDescription0",
                    ]
            else:
                ixn_cpdpconvergence.Enabled = False

        except Exception as err:
            raise SnappiIxnException(err)

        bad_requests = self.get_json_import_errors()
        if bad_requests != []:
            bad_requests.insert(0, "bad request errors from Ixn:")
            raise SnappiIxnException(400, bad_requests)
        return self._request_detail()

    def get_json_import_errors(self):
        app_errors = self._globals.AppErrors.find()
        bad_requests = []
        if len(app_errors) > 0:
            current_errors = app_errors[0].Error.find()
            if len(current_errors) > 0:
                for error in current_errors:
                    if error.Name != "JSON Import Issues":
                        continue
                    for instance in error.Instance.find():
                        if instance.SourceValues in self._previous_errors:
                            continue
                        bad_requests.append(instance.SourceValues[0])
                        # this is for linux as it is retaining errors from
                        # previous run and raising exception in the latest
                        # script run
                        self._previous_errors.append(instance.SourceValues)
        return bad_requests

    def config_ixnetwork(self, config):
        self._config_objects = {}
        self._device_encap = {}
        self._device_traffic_endpoint = {}
        self.ixn_objects = IxNetObjects(self)
        self.ixn_routes = IxNetObjects(self)
        self._dev_compacted = {}
        self._connect()
        self.capture.reset_capture_request()
        self._config = self._validate_instance(config)
        with Timer(self, "Config validation"):
            self.validation.validate_config()
        self._ixnetwork.Traffic.UseRfc5952 = True
        if len(self._config._properties) == 0:
            self._ixnetwork.NewConfig()
        else:
            self.vport.config()
            self.lag.config()
            with Timer(self, "Devices configuration"):
                self.ngpf.config()
            self.traffic_item.config()
        self._running_config = self._config
        self._apply_change()
        with Timer(self, "Start interfaces"):
            # Start all protocols is workaround for pCPU crash reported by
            # Microsoft, need to revert once fix is available for pCPU crash
            if self._protocols_exists():
                for device in self._config.devices:
                    # After the config is pushed in IxNetwork, only then we can do ClearOverlays() in MKA Globals.
                    # That is why it cannot be done from within MKA module itself.

                    # Check if MACsec/MKA exists in topology
                    macsec = device.get("macsec")
                    if macsec:
                        if self._macsec._is_dynamic_key(macsec):
                            # it is MKA
                            self._mka._clear_overlays_in_globals(macsec)
                self._start_interface()
            else:
                if len(self._ixnetwork.Topology.find()) > 0:
                    self._ixnetwork.StartAllProtocols()
        if len(self.lag._lags_config) > 0:
            for lag in self.lag._lags_config:
                if lag.min_links > len(lag.ports):
                    self.warning(
                        "ports in {0} are less than configured minimum links {1} so {0} is inactive ".format(
                            lag.name, lag.min_links
                        )
                    )
                    self._ixnetwork.Lag.find(Name=lag.name).Stop()

    def _protocols_exists(self):
        total_dev = len(self._ixnetwork.GetTopologyStatus())
        topos = self._ixnetwork.Topology.find()
        ethv4v6_dev_count = 0
        if len(topos) > 0:
            dgs = topos.DeviceGroup.find()
            if len(dgs) > 0:
                eth_dev = len(
                    self._ixnetwork.Topology.find()
                    .DeviceGroup.find()
                    .Ethernet.find()
                )
                ethv4v6_dev_count = ethv4v6_dev_count + eth_dev
                if eth_dev > 0:
                    v4_dev = len(
                        self._ixnetwork.Topology.find()
                        .DeviceGroup.find()
                        .Ethernet.find()
                        .Ipv4.find()
                    )
                    ethv4v6_dev_count = ethv4v6_dev_count + v4_dev
                    v6_dev = len(
                        self._ixnetwork.Topology.find()
                        .DeviceGroup.find()
                        .Ethernet.find()
                        .Ipv6.find()
                    )
                    ethv4v6_dev_count = ethv4v6_dev_count + v6_dev
        if total_dev > ethv4v6_dev_count:
            return True
        else:
            return False

    def set_control_state(self, payload):
        try:
            control_option = payload.choice
            control_obj = getattr(payload, control_option)
            control_choice = control_obj.get("choice")
            request_payload = getattr(control_obj, control_choice)
            self._connect()
            event_names = []
            event_state = None
            event_type = control_choice
            EventInfo = namedtuple(
                "EventInfo", ["event_type", "event_state", "event_names"]
            )
            if control_option == "port":
                if control_choice == "capture":
                    self.capture.set_capture_state(request_payload)
                elif control_choice == "link":
                    self._TRIGGERED_EVENT = "link"
                    if request_payload.port_names is not None:
                        event_names = request_payload.port_names
                        event_state = request_payload.state
                    self.vport.set_link_state(request_payload)
            elif control_option == "protocol":
                if control_choice == "all":
                    self.ngpf.set_protocol_state(request_payload)
                elif control_choice == "route":
                    event_state = request_payload.state
                    with Timer(self, "Setting route state"):
                        event_names = self.ngpf.set_route_state(
                            request_payload
                        )
                elif control_choice == "lacp":
                    self.ngpf.set_device_state(request_payload)
            elif control_option == "traffic":
                self.traffic_item.transmit(request_payload)
            elif control_option is not None:
                msg = "{} is not a supported choice for metrics; \
                the supported choices are \
                ['port', 'protocol', traffic]".format(
                    control_option
                )
                raise SnappiIxnException(400, msg)
            self._event_info = EventInfo(event_type, event_state, event_names)
        except Exception as err:
            raise SnappiIxnException(err)
        return self._request_detail()

    def set_control_action(self, payload):
        try:
            control_option = payload.choice
            control_obj = getattr(payload, control_option)
            control_choice = control_obj.get("choice")
            choice_obj = getattr(control_obj, control_choice)
            if control_choice == "ipv4":
                choice = choice_obj.get("choice")
                request_payload = getattr(choice_obj, choice)
                if choice == "ping":
                    res = self.control_action_response()
                    self._connect()
                    res.response.protocol.ipv4.ping.responses.deserialize(
                        self.ping.results(request_payload, control_choice)
                    )
            elif control_choice == "ipv6":
                choice = choice_obj.get("choice")
                request_payload = getattr(choice_obj, choice)
                if choice == "ping":
                    res = self.control_action_response()
                    self._connect()
                    res.response.protocol.ipv6.ping.responses.deserialize(
                        self.ping.results(request_payload, control_choice)
                    )
            elif control_option is not None:
                msg = "{} is not a supported choice for metrics; \
                the supported choices are \
                ['ipv4', 'ipv6']".format(
                    control_option
                )
                raise SnappiIxnException(400, msg)
            res.warnings = snappi.Warning()
            return res
        except Exception as err:
            raise SnappiIxnException(err)

    def get_capture(self, request):
        """Gets capture file and returns it as a byte stream"""
        try:
            if (
                isinstance(request, (type(self._capture_request), str))
                is False
            ):
                raise TypeError(
                    "The content must be of type Union[CaptureRequest, str]"
                )
            if isinstance(request, str) is True:
                request = self._capture_request.deserialize(request)
            self._connect()
        except Exception as err:
            raise SnappiIxnException(err)
        return self.capture.results(request)

    def get_states(self, request):
        try:
            states_request = self.states_request()
            if isinstance(request, (type(states_request), str)) is False:
                raise TypeError(
                    "The content must be of type Union[StatesRequest, str]"
                )
            if isinstance(request, str) is True:
                request = states_request.deserialize(request)
            self._connect()
            response = self.ngpf.get_states(request)
            states_response = self.states_response()
            states_response.deserialize(response)
            return states_response
        except Exception as err:
            raise SnappiIxnException(err)

    def get_metrics(self, request):
        """
        Gets port, flow and protocol metrics.

        Args
        ----
        - request (Union[MetricsRequest, str]): A request for Port, Flow and
          protocol metrics.
          The request content MUST be vase on the OpenAPI model,
          #/components/schemas/Result.MetricsRequest
          See the docs/openapi.yaml document for all model details
        """
        try:
            self._connect()
            metric_req = self.metrics_request()
            if isinstance(request, (type(metric_req), str)) is False:
                raise TypeError(
                    "The content must be of type Union[MetricsRequest, str]"
                )
            if isinstance(request, str) is True:
                request = metric_req.deserialize(request)
            # Need to change the code style when the choice Enum grows big
            if request.get("choice") == "port":
                response = self.vport.results(request.port)
                metric_res = self.metrics_response()
                metric_res.port_metrics.deserialize(response)
                return metric_res
            if request.get("choice") == "flow":
                response = self.traffic_item.results(request.flow)
                metric_res = self.metrics_response()
                metric_res.flow_metrics.deserialize(response)
                return metric_res
            if request.get("choice") == "egress_only_tracking":
                response = self.traffic_item.results_egress_only_tracking(
                    request.egress_only_tracking
                )
                metric_res = self.metrics_response()
                metric_res.egress_only_tracking_metrics.deserialize(response)
                return metric_res
            if request.get("choice") == "lag":
                response = self.traffic_item.results(request.lag)
                metric_res = self.metrics_response()
                metric_res.lag_metrics.deserialize(response)
                return metric_res
            if request.get("choice") == "lacp":
                response = self.traffic_item.results(request.lacp)
                metric_res = self.metrics_response()
                metric_res.lacp_metrics.deserialize(response)
                return metric_res
            if request.get("choice") == "convergence":
                response = self._result(request.convergence)
                metric_res = self.metrics_response()
                metric_res.convergence_metrics.deserialize(response)
                return metric_res
            if request.get("choice") == "rocev2_flow":
                response = self.traffic_item.rocev2_flow_results(
                    request.rocev2_flow
                )  # noqa
                metric_res = self.metrics_response()
                metric_res.rocev2_flow_per_qp_metrics.deserialize(response)
                return metric_res
            if (
                request.get("choice")
                in self.protocol_metrics.get_supported_protocols()
            ):
                response = self.protocol_metrics.results(request)
                metric_res = self.metrics_response()
                if (
                    request.choice == "rocev2_ipv4"
                    or request.choice == "rocev2_ipv6"
                ):
                    getattr(
                        metric_res, request.choice + "_per_peer" + "_metrics"
                    ).deserialize(response)
                else:
                    getattr(
                        metric_res, request.choice + "_metrics"
                    ).deserialize(response)
                return metric_res
        except Exception as err:
            raise SnappiIxnException(err)
        if request.get("choice") is not None:
            msg = "{} is not a supported choice for metrics; \
            the supported choices are \
            ['port', 'flow']".format(
                request.choice
            )
            raise SnappiIxnException(400, msg)

    def update_flows(self, payload):
        """
        Update Flows for property size and rate

        Args
        ----
        - request (Union[UpdateFlows, str]): A request for Flow property update.
          The request content MUST be vase on the OpenAPI model,
          #/components/schemas/Control.FlowsUpdate
          See the docs/openapi.yaml document for all model details
        """
        try:
            if isinstance(payload, str) is True:
                payload = self._flows_update.deserialize(payload)
            self._connect()
            self.traffic_item.update_flows(payload)
        except Exception as err:
            raise SnappiIxnException(err)
        return self._request_detail()

    def delete_config(self, payload):
        """
        Delete Flows from config
        Args
        ----
        - request (Union[ConfigDeleteResources, str]): A request for Flow name for deletion.  # noqa
          The request content MUST be vase on the OpenAPI model,
          #/components/schemas/Config.Delete
          See the docs/openapi.yaml document for all model details
        """
        try:
            if isinstance(payload, str) is True:
                payload = self._flows_delete.deserialize(payload)
            self._connect()
            self.traffic_item.delete_configs(payload)
        except Exception as err:
            raise SnappiIxnException(err)
        return self._request_detail()

    def append_config(self, payload):
        """
        Append Flows for config
        Args
        ----
        - request (Union[ConfigAppendResources, str]): A request for Flow name for append.  # noqa
          The request content MUST be vase on the OpenAPI model,
          #/components/schemas/Config.Append
          See the docs/openapi.yaml document for all model details
        """
        try:
            if isinstance(payload, str) is True:
                payload = self._flows_append.deserialize(payload)
            self._connect()
            self.traffic_item.append_configs(payload)
        except Exception as err:
            raise SnappiIxnException(err)
        return self._request_detail()

    def _get_max_convergence(self, rows):
        # We are extracting max value for multiple destination
        rows_len = len(rows)
        if rows_len > 1:
            for (
                _,
                internal_name,
                external_type,
            ) in self._CONVERGENCE:
                if external_type not in ["float", "int"]:
                    continue
                max_value = rows[0][internal_name]
                for idx in range(1, rows_len):
                    tmp_value = rows[idx][internal_name]
                    if max_value < tmp_value:
                        max_value = tmp_value
                rows[0][internal_name] = max_value
        return rows[0]

    def _result(self, request):
        # get convergence result
        flow_names = request.get("flow_names")
        if not isinstance(flow_names, list):
            msg = "Invalid format of flow_names passed {},\
                                                expected list".format(
                flow_names
            )
            raise Exception(msg)
        if flow_names is None or len(flow_names) == 0:
            flow_names = [flow.name for flow in self.snappi_config.flows]
        flow_rows = self._get_flow_rows(flow_names)
        traffic_stat = self._assistant.StatViewAssistant(
            "Traffic Item Statistics"
        )
        traffic_index = {}
        drill_down_option = "Drill down per Dest Endpoint"
        for index, row in enumerate(
            self._get_traffic_rows(traffic_stat, drill_down_option)
        ):
            traffic_index[row["Traffic Item"]] = index
        response = []
        for flow_name in flow_names:
            traffic_stat = self._assistant.StatViewAssistant(
                "Traffic Item Statistics"
            )
            convergence = {"name": flow_name}
            if flow_name not in traffic_index.keys():
                raise Exception("Somehow flow %s is missing" % flow_name)

            drill_down_options = traffic_stat.DrillDownOptions()
            drilldown_index = drill_down_options.index(drill_down_option)
            drill_down = traffic_stat.Drilldown(
                drilldown_index,
                drill_down_option,
                traffic_stat.TargetRowFilters()[traffic_index[flow_name]],
            )
            if drill_down.Rows is not None:
                drill_down_result = self._get_max_convergence(drill_down.Rows)
            ixn_cpdpconvergence = self._traffic.Statistics.CpdpConvergence
            if (
                ixn_cpdpconvergence.EnableDataPlaneEventsRateMonitor
                and ixn_cpdpconvergence.EnableControlPlaneEvents
            ):  # noqa
                for (
                    external_name,
                    internal_name,
                    external_type,
                ) in self._CONVERGENCE:
                    self._set_result_value(
                        convergence,
                        external_name,
                        drill_down_result[internal_name],
                        external_type,
                    )
                events = []
                interruption_time = None
                for flow_result in flow_rows:
                    if flow_result["Traffic Item"] != flow_name:
                        continue
                    event_name = flow_result["Event Name"]
                    if event_name == "":
                        continue
                    event = self._get_event(event_name)
                    for (
                        external_name,
                        internal_name,
                        external_type,
                    ) in self._EVENT:
                        value = int(flow_result[internal_name].split(".")[-1])
                        self._set_result_value(
                            event, external_name, value * 1000, external_type
                        )
                    events.append(event)
                    if interruption_time is None:
                        if flow_result["DP Above Threshold Timestamp"] != "":
                            att = flow_result[
                                "DP Above Threshold Timestamp"
                            ].split(":")[-1]
                        else:
                            att = 0
                        if flow_result["DP Below Threshold Timestamp"] != "":
                            btt = flow_result[
                                "DP Below Threshold Timestamp"
                            ].split(":")[-1]
                        else:
                            btt = 0
                        interruption_time = float(att) - float(btt)
                        self._set_result_value(
                            convergence,
                            "service_interruption_time_us",
                            interruption_time,
                            float,
                        )
                convergence["events"] = events

            # for DP only metric
            if (
                ixn_cpdpconvergence.EnableDataPlaneEventsRateMonitor
                and not ixn_cpdpconvergence.EnableControlPlaneEvents
            ):  # noqa
                for (
                    external_name,
                    internal_name,
                    external_type,
                ) in self._DP_CONVERGENCE:
                    self._set_result_value(
                        convergence,
                        external_name,
                        drill_down_result[internal_name],
                        external_type,
                    )
            response.append(convergence)
        return response

    def _get_event(self, event_name):
        event = {}
        if re.search(r"Port Link Up", event_name):
            if self._event_info.event_state == "up":
                event["type"] = "link_up"
            else:
                event["type"] = "link_down"
        else:
            for route_name in self.ixn_routes.names:
                if re.search(route_name, event_name) is not None:
                    event["source"] = route_name
                    event_type = event_name.split(route_name)[-1]
                    if event_type.strip().lower() == "disable":
                        event_type = "route_withdraw"
                    else:
                        event_type = "route_advertise"
                    event["type"] = event_type
                    break
        return event

    def _set_result_value(
        self, row, column_name, column_value, column_type=str
    ):
        try:
            row[column_name] = column_type(column_value)
        except:
            if column_type.__name__ in ["float", "int"]:
                row[column_name] = 0
            else:
                row[column_type] = column_value

    def _get_traffic_rows(self, traffic_stat, drill_down_option):
        count = 0
        sleep_time = 0.5
        while True:
            drill_down_options = traffic_stat.DrillDownOptions()
            if drill_down_option in drill_down_options:
                break
            if count * sleep_time > self._convergence_timeout:
                raise Exception(
                    "Somehow 'Drill down per Dest Endpoint' not available"
                )
            time.sleep(sleep_time)
            count += 1
        return traffic_stat.Rows

    def _get_flow_rows(self, flow_names):
        count = 0
        sleep_time = 0.5
        flow_stat = self.assistant.StatViewAssistant("Flow Statistics")

        if self._TRIGGERED_EVENT == "link":
            if "Show Physical Ports in LAG" in flow_stat.DrillDownOptions():
                flow_index = {}
                drill_down_option = "Show Physical Ports in LAG"
                for index, row in enumerate(
                    self._get_traffic_rows(flow_stat, drill_down_option)
                ):
                    flow_index[row["Traffic Item"]] = index
                drill_down_options = flow_stat.DrillDownOptions()
                drilldown_index = drill_down_options.index(drill_down_option)
                flow_stat.Drilldown(
                    0,
                    drill_down_option,
                    flow_stat.TargetRowFilters()[drilldown_index],
                )
                flow_stat = self.assistant.StatViewAssistant("Flow Statistics")
        has_flow = False
        while True:
            flow_rows = flow_stat.Rows
            has_event = False
            ixn_cpdpconvergence = self._traffic.Statistics.CpdpConvergence
            for row in flow_rows:
                if row["Traffic Item"] in flow_names:
                    has_flow = True
                    if (
                        ixn_cpdpconvergence.EnableDataPlaneEventsRateMonitor
                        and ixn_cpdpconvergence.EnableControlPlaneEvents
                    ):  # noqa
                        if row["Event Name"] != "":
                            has_event = True
                            break
            if has_event is True:
                break
            if count * sleep_time > self._convergence_timeout:
                if has_flow is not True:
                    raise Exception(
                        "flow_names must present within in config.flows"
                    )
                else:
                    self.info("event is not reflected in stat")
                    break
            time.sleep(sleep_time)
            count += 1
        return flow_rows

    def add_error(self, error):
        """Add an error to the global errors"""
        if isinstance(error, str) is False:
            self._errors.append("%s %s" % (type(error), str(error)))
        else:
            self._errors.append(error)

    def get_errors(self):
        return self._errors

    def parse_location_info(self, location):
        """It will return (chassis,card,port)
        set card as 0 where that is not applicable"""
        if re.search("/|;", location) is None:
            raise Exception(
                "Invalid port location format, expected ["
                "chassis_ip;card_id;port_id or chassis_ip/port_id]"
            )
        LocationInfo = namedtuple(
            "LocationInfo", ["chassis_info", "card_info", "port_info"]
        )
        if ";" in location:
            try:
                (chassis_info, card_info, port_info) = location.split(";")
            except Exception:
                raise Exception("Please specify <chassis>;<card>;<port>")
        else:
            try:
                card_info = 0
                (chassis_info, port_info) = location.split("/")
            except Exception:
                raise Exception("Please specify <chassis>/<port>")
        return LocationInfo(chassis_info, card_info, port_info)

    def special_char(self, names):
        is_names = True
        if not isinstance(names, list):
            is_names = False
            names = [names]

        ret_list = []
        for name in names:
            if name is None:
                ret_list.append(name)
            else:
                ret_list.append(
                    name.replace("(", "\\(")
                    .replace(")", "\\)")
                    .replace("[", "\\[")
                    .replace("]", "\\]")
                    .replace(".", "\\.")
                    .replace("*", "\\*")
                    .replace("+", "\\+")
                    .replace("?", "\\?")
                    .replace("{", "\\{")
                    .replace("}", "\\}")
                )
        if is_names is True:
            return ret_list
        else:
            return ret_list[0]

    def _get_restpy_trace(self, log_level):
        if log_level == logging.DEBUG:
            return SessionAssistant.LOGLEVEL_ALL
        elif log_level == logging.WARNING:
            return SessionAssistant.LOGLEVEL_WARNING
        return SessionAssistant.LOGLEVEL_INFO

    def _connect(self):
        """Connect to an IxNetwork API Server."""
        self._errors = []
        self.logger = setup_ixnet_logger(self.log_level, module_name=__name__)
        if self._assistant is None:
            platform = TestPlatform(self._address, rest_port=self._port)
            platform.Authenticate(self._username, self._password)
            url = "%s://%s:%s/ixnetworkweb/api/v1/usersettings/ixnrest" % (
                platform.Scheme,
                platform.Hostname,
                platform.RestPort,
            )
            platform._connection._session.request(
                "put",
                url,
                data=json.dumps({"enableClassicProtocols": True}),
                verify=False,
            )
            self._assistant = SessionAssistant(
                IpAddress=self._address,
                RestPort=self._port,
                UserName=self._username,
                Password=self._password,
                LogLevel=self._get_restpy_trace(self._log_level),
            )
            self._ixnetwork = self._assistant.Session.Ixnetwork
            self._vport = self._ixnetwork.Vport
            self._lag = self._ixnetwork.Lag
            self._topology = self._ixnetwork.Topology
            self._traffic = self._ixnetwork.Traffic
            self._traffic_item = self._ixnetwork.Traffic.TrafficItem
            self._globals = self._ixnetwork.Globals
            if not self._ixn_version_check():
                raise Exception(
                    "IxNetwork 9.10 or newer is required for snappi[ixnetwork]"
                )
            if len(self._license_servers) > 0:
                self._ixnetwork.Globals.Licensing.LicensingServers = (
                    self._license_servers
                )
            try:
                import pkg_resources

                snappi_ver = (
                    "snappi-"
                    + pkg_resources.get_distribution("snappi").version
                )
                self.info(snappi_ver)
                snappi_ixn = (
                    "snappi_ixnetwork-"
                    + pkg_resources.get_distribution(
                        "snappi_ixnetwork"
                    ).version
                )
                self.info(snappi_ixn)
                restpy = (
                    "ixnetwork_restpy-"
                    + pkg_resources.get_distribution(
                        "ixnetwork_restpy"
                    ).version
                )
                self.info(restpy)
            except pkg_resources.DistributionNotFound as e:
                version = "Could not determine version for pkg {}".format(
                    e.req.project_name
                )
                self.info(version)
            except Exception as e:
                self.warning("{}".format(e))
        self._backup_errors()

    def _ixn_version_check(self):
        major, minor = self._globals.BuildNumber.split(".")[0:2]
        if int(major) < 9:
            return False
        if int(major) == 9 and int(minor) < 10:
            return False
        return True

    def _backup_errors(self):
        app_errors = self._globals.AppErrors.find()
        if len(app_errors) > 0:
            self._ixn_errors = app_errors[0].Error.find()

    def _validate_instance(self, config):
        """Validate current IxNetwork instance:
        1. Stop everything if local config is None
        2. Otherwise add warning message"""
        traffic_state = self._traffic.State
        if self.snappi_config is None:
            if traffic_state == "started":
                self._traffic_item.find()
                if len(self._traffic_item) > 0:
                    self._traffic_item.StopStatelessTrafficBlocking()
            glob_topo = self._globals.Topology.refresh()
            if glob_topo.Status == "started":
                self._ixnetwork.StopAllProtocols("sync")
        else:
            if traffic_state == "started":
                msg = (
                    "Flows are in running state. "
                    "Please stop those using set_transmit_state"
                )
                self.add_error(msg)
                self.warning(msg)
        self._initial_flows_config = self.config().flows.deserialize(
            config.flows.serialize(config.flows.DICT)
        )

        if "UHD" in self._ixnetwork.Globals.ProductVersion:
            chassis_info = "localuhd"
            for port in config.ports:
                if port.location is None:
                    continue
                if ";" in port.location:
                    (_, card, port_info) = port.location.split(";")
                    port_info = "{}.{}".format(card, port_info)
                elif "/" in port.location:
                    (_, port_info) = port.location.split("/")
                else:
                    raise SnappiIxnException(400, "port location is not valid")
                port.location = chassis_info + "/" + port_info
        return config

    def _apply_change(self):
        """Apply on the fly only applicable for Device object"""
        glob_topo = self._globals.Topology.refresh()
        if glob_topo.ApplyOnTheFlyState == "allowed":
            url = (
                "%s/globals/topology/operations/applyonthefly"
                % self._ixnetwork.href
            )
            payload = {"arg1": glob_topo.href}
            # todo: Sometime it redirect to some unknown loaction
            try:
                self._request("POST", url, payload)
            except Exception:
                pass

    def _start_interface(self):
        topos = self._ixnetwork.Topology.find()
        if len(topos) > 0:
            dgs = topos.DeviceGroup.find()
            if len(dgs) > 0:
                eth_list = dgs.Ethernet.find()
                if len(eth_list) > 0:
                    ip4_list = eth_list.Ipv4.find()
                    ip6_list = eth_list.Ipv6.find()
                    if len(eth_list) == max(len(ip4_list), len(ip6_list)):
                        if len(ip4_list) > 0:
                            ip4_list.Start()
                        if len(ip6_list) > 0:
                            ip6_list.Start()
                    else:
                        eth_list.Start()
                        if len(ip4_list) > 0:
                            ip4_list.Start()
                        if len(ip6_list) > 0:
                            ip6_list.Start()

    def _request(self, method, url, payload=None):
        self.debug("Request and Response ...")
        connection, url = self._assistant.Session._connection._normalize_url(
            url
        )
        headers = {
            "Content-Type": "application/json",
            "x-api-key": self._assistant.Session._connection.x_api_key,
        }
        if payload is not None:
            payload = json.dumps(payload)
        self.debug("%s %s %s" % (method, url, payload))
        response = self._assistant.Session._connection._session.request(
            method, url, headers=headers, data=payload, verify=False
        )
        response.raise_for_status()
        self.debug("Response %s" % response)
        if response.status_code == 202:
            content = response.json()
            while content["state"] == "IN_PROGRESS":
                time.sleep(0.2)
                response = self._request("GET", content["url"])
        if response.headers.get("Content-Type"):
            if response.headers["Content-Type"] == "application/json":
                return response.json()
            elif (
                response.headers["Content-Type"] == "application/octet-stream"
            ):
                return response.content
        return response

    def _remove(self, ixn_obj, items):
        """Remove any ixnetwork objects that are not found in the items list.
        If the items list does not exist remove everything.
        """
        valid_names = []
        for item in items:
            if isinstance(item, dict):
                name = item.get("name")
            else:
                name = item.name
            if name is not None:
                valid_names.append(name)
        invalid_names = []
        for item in ixn_obj.find():
            if item.Name not in valid_names:
                invalid_names.append(item.Name)
        self.debug("Removing these %s" % invalid_names)
        if len(invalid_names) > 0:
            if ixn_obj._SDM_NAME == "trafficItem":
                # can't remove traffic items that are started
                start_states = [
                    "txStopWatchExpected",
                    "locked",
                    "started",
                    "startedWaitingForStats",
                    "startedWaitingForStreams",
                    "stoppedWaitingForStats",
                ]
                for item in ixn_obj.find(
                    Name="^(%s)$" % "|".join(self.special_char(invalid_names))
                ):
                    if item.State in start_states:
                        item.StopStatelessTraffic()
                if len(ixn_obj) > 0:
                    poll = True
                    while poll:
                        poll = False
                        for v in self.select_traffic_items().values():
                            if v["state"] not in [
                                "error",
                                "stopped",
                                "unapplied",
                            ]:
                                poll = True
            ixn_obj.find(
                Name="^(%s)$" % "|".join(self.special_char(invalid_names))
            )
            if len(ixn_obj) > 0:
                ixn_obj.remove()

    # def _get_topology_name(self, port_name):
    #     return "Topology %s" % port_name

    def select_card_aggregation(self, location):
        (hostname, cardid, portid) = location.split(";")
        payload = {
            "selects": [
                {
                    "from": "/availableHardware",
                    "properties": [],
                    "children": [
                        {
                            "child": "chassis",
                            "properties": [],
                            "filters": [
                                {
                                    "property": "hostname",
                                    "regex": "^%s$" % hostname,
                                }
                            ],
                        },
                        {
                            "child": "card",
                            "properties": ["*"],
                            "filters": [
                                {
                                    "property": "cardId",
                                    "regex": "^%s$" % abs(int(cardid)),
                                }
                            ],
                        },
                        {
                            "child": "aggregation",
                            "properties": ["*"],
                            "filters": [],
                        },
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self._ixnetwork.href
        results = self._ixnetwork._connection._execute(url, payload)
        return results[0]["chassis"][0]["card"][0]

    def select_chassis_card(self, vport):
        pieces = vport["connectionStatus"].split(";")
        payload = {
            "selects": [
                {
                    "from": "/availableHardware",
                    "properties": [],
                    "children": [
                        {
                            "child": "chassis",
                            "properties": [],
                            "filters": [
                                {
                                    "property": "hostname",
                                    "regex": "^%s$" % pieces[0],
                                }
                            ],
                        },
                        {
                            "child": "card",
                            "properties": ["*"],
                            "filters": [
                                {
                                    "property": "cardId",
                                    "regex": "^%s$" % int(pieces[1]),
                                }
                            ],
                        },
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self._ixnetwork.href
        results = self._ixnetwork._connection._execute(url, payload)
        return results[0]["chassis"][0]["card"][0]

    def select_vports(self, port_name_filters=[]):
        """Select all vports.
        Return them in a dict keyed by vport name.
        """
        payload = {
            "selects": [
                {
                    "from": "/",
                    "properties": [],
                    "children": [
                        {
                            "child": "vport",
                            "properties": [
                                "name",
                                "type",
                                "location",
                                "connectionState",
                                "connectionStatus",
                                "assignedTo",
                                "connectedTo",
                            ],
                            "filters": port_name_filters,
                        },
                        {
                            "child": "l1Config",
                            "properties": ["currentType"],
                            "filters": [],
                        },
                        {
                            "child": "capture",
                            "properties": [
                                "hardwareEnabled",
                                "softwareEnabled",
                            ],
                            "filters": [],
                        },
                        {
                            "child": "^(eth.*|novus.*|uhd.*|atlas.*|ares.*|star.*|ten.*)$",
                            "properties": ["*"],
                            "filters": [],
                        },
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self._ixnetwork.href
        results = self._ixnetwork._connection._execute(url, payload)
        vports = {}
        if "vport" in results[0]:
            for vport in results[0]["vport"]:
                vports[vport["name"]] = vport
        return vports

    def select_traffic_items(self, traffic_item_filters=[]):
        """Select all traffic items.
        Return them in a dict keyed by traffic item name.

        Args
        ----
        - filters (list(dict(property:'', 'regex':''))): A list of filters for the select.
            A filter is a dict with a property name and a regex match
        """
        payload = {
            "selects": [
                {
                    "from": "/traffic",
                    "properties": [],
                    "children": [
                        {
                            "child": "trafficItem",
                            "properties": ["name", "state", "enabled"],
                            "filters": traffic_item_filters,
                        },
                        {
                            "child": "highLevelStream",
                            "properties": [
                                "txPortName",
                                "rxPortNames",
                                "state",
                                "name",
                            ],
                            "filters": [],
                        },
                        {
                            "child": "tracking",
                            "properties": ["trackBy"],
                            "filters": [],
                        },
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self._ixnetwork.href
        results = self._ixnetwork._connection._execute(url, payload)
        traffic_items = {}
        try:
            for traffic_item in results[0]["trafficItem"]:
                traffic_items[traffic_item["name"]] = traffic_item
        except Exception:
            pass
        return traffic_items

    def select_chassis_card_port(self, location):
        """Select all availalehardware.
        Return them in a dict keyed by vport name.
        """
        (hostname, cardid, portid) = location.split(";")
        payload = {
            "selects": [
                {
                    "from": "/availableHardware",
                    "properties": [],
                    "children": [
                        {
                            "child": "chassis",
                            "properties": [],
                            "filters": [
                                {
                                    "property": "hostname",
                                    "regex": "^%s$" % hostname,
                                }
                            ],
                        },
                        {
                            "child": "card",
                            "properties": [],
                            "filters": [
                                {
                                    "property": "cardId",
                                    "regex": "^%s$" % abs(int(cardid)),
                                }
                            ],
                        },
                        {
                            "child": "port",
                            "properties": [],
                            "filters": [
                                {
                                    "property": "portId",
                                    "regex": "^%s$" % abs(int(portid)),
                                }
                            ],
                        },
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self._ixnetwork.href
        results = self._ixnetwork._connection._execute(url, payload)
        return results[0]["chassis"][0]["card"][0]["port"][0]["xpath"]

    def clear_ownership(self, available_hardware_hrefs, location_hrefs):
        self.debug(
            "Clearing ownership %s from %s"
            % (available_hardware_hrefs, location_hrefs)
        )
        hrefs = list(available_hardware_hrefs.values()) + list(
            location_hrefs.values()
        )
        if len(hrefs) == 0:
            return
        names = list(available_hardware_hrefs.keys()) + list(
            location_hrefs.keys()
        )
        with Timer(self, "Location preemption [%s]" % ", ".join(names)):
            payload = {"arg1": [href for href in hrefs]}
            url = "%s/operations/clearownership" % payload["arg1"][0]
            self._ixnetwork._connection._execute(url, payload)

    def get_config(self):
        return self._config

    def check_protocol_statistics(self):
        start = time.time()
        url = "%s/operations/gettopologystatus" % self._ixnetwork.href
        check = True
        while check is True and time.time() - start < 90:
            check = False
            results = self._ixnetwork._connection._execute(url, None)
            for result in results:
                if result["arg2"][0]["arg2"] != result["arg2"][3]["arg2"]:
                    check = True

    def info(self, message):
        self.logger.info(message)

    def debug(self, message):
        self.logger.debug(message)

    def warning(self, message):
        logging.warning(message)

    def get_version(self):
        try:
            import pkg_resources

            sdk_version = (
                "snappi-" + pkg_resources.get_distribution("snappi").version
            )
            app_version = (
                "snappi_ixnetwork-"
                + pkg_resources.get_distribution("snappi_ixnetwork").version
            )

            return {
                "api_spec_version": "open-api-models-"
                + snappi.Api.get_local_version(self).api_spec_version,
                "sdk_version": sdk_version,
                "app_version": app_version,
            }
        except:
            raise SnappiIxnException("unable to get version")



from copy import deepcopy


class CustomField(object):
    """
    Implemented All custom field which specify within TrafficItem class
    Best Practice :
        - Please use self._configure_pattern to setting Pattern
    """

    class DummyPattern(object):
        def get(self, value):
            return getattr(self, value)

    @classmethod
    def _process_ipv4_priority(
        cls,
        tr_instance,
        ixn_fields,
        field_names,
        snappi_header,
        priority,
        stacks=None,
    ):
        priority = snappi_header.get(priority, True)
        choice = priority.get("choice")
        field_map = tr_instance._IPV4
        priority_obj = priority.get(choice, True)
        phb = None
        prop_types = sorted(priority_obj._TYPES)
        prop_types.reverse()
        for field in prop_types:
            property = priority_obj.get(field, True)
            if field == "phb":
                phb = property
                field = tr_instance._get_phb_type(property)
            if field == "ecn":
                field = tr_instance._get_phb_type(phb)
                field = "{}-unused".format(field)
            try:
                ind = field_names.index(field_map[field])
            except Exception:
                continue
            tr_instance._config_field_pattern(
                property, ixn_fields[ind], None, True
            )

    @classmethod
    def _process_ethernet_pause(
        cls,
        tr_instance,
        ixn_fields,
        field_names,
        snappi_header,
        ethpause,
        stacks=None,
    ):
        xpath = ixn_fields[0]["xpath"].split("/stack[@alias")[0]
        xpath = "{CE}/stack[@alias = '{HEADER}-{INDEX}']".format(
            CE=xpath, HEADER="custom", INDEX=2
        )
        custom = snappi_header.parent.custom
        control_op_code = tr_instance._get_first_value(
            snappi_header.get("control_op_code", True)
        )

        time = tr_instance._get_first_value(snappi_header.get("time", True))

        custom.bytes = "{:04x}{:x}".format(control_op_code, time)
        tr_instance._append_header(xpath, stacks, custom)

    @classmethod
    def _process_custom_header(
        cls,
        tr_instance,
        ixn_fields,
        field_names,
        snappi_header,
        custom,
        stacks=None,
    ):
        custom_bytes = snappi_header.bytes
        field_map = tr_instance._CUSTOM
        pattern = tr_instance.DummyPattern()
        pattern.choice = "value"
        pattern.value = len(custom_bytes) * 4
        ind = field_names.index(field_map["length"])
        tr_instance._config_field_pattern(pattern, ixn_fields[ind])
        pattern.value = custom_bytes
        ind = field_names.index(field_map["data"])
        tr_instance._config_field_pattern(pattern, ixn_fields[ind])

    def _get_first_value(self, pattern):
        if pattern.choice is None:
            return None
        if pattern.choice == "value":
            value = pattern.value
        elif pattern.choice == "values":
            value = pattern.values[0]
        elif pattern.choice == "counter":
            value = pattern.counter.start
        else:
            value = pattern.random.min
        return value

    def _get_phb_type(self, phb_pattern):
        value = self._get_first_value(phb_pattern)
        phb_type = "default"
        if value is None:
            value = 0
        if value > 0:
            phb_type = "phb" if value % 8 == 0 else "af"
        if value == 46:
            phb_type = "ef"
        return phb_type

    def _gtpv1(self, new_headers, header):
        """"""
        new_headers.append(header)
        gtp_option = deepcopy(header)
        gtp_option.__setattr__("choice", "gtpv1option")
        gtp_option.__setattr__("gtpv1option", header.gtpv1)
        gtp_option.__delattr__("gtpv1")
        new_headers.append(gtp_option)



def test_clear_config(api):
    """Demonstrates how to clear an IxNetwork session configuration"""
    api.set_config(api.config())
    assert len(api._assistant.Ixnetwork.Vport.find()) == 0
    assert len(api._assistant.Ixnetwork.Topology.find()) == 0
    assert len(api._assistant.Ixnetwork.Traffic.TrafficItem.find()) == 0



from snappi_ixnetwork.logger import get_ixnet_logger

__all__ = ["Base", "MultiValue", "PostCalculated"]


class MultiValue(object):
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value


class PostCalculated(object):
    def __init__(self, key, ref_ixnobj=None, ixnobj=None):
        self._key = key
        self._ref_obj = ref_ixnobj
        self._parent_obj = ixnobj
        self.logger = get_ixnet_logger(__name__)

    @property
    def value(self):
        value = None
        if self._key == "connectedTo":
            value = self._ref_obj.get("xpath")
        self.logger.debug("Post Calculated %s - %s" % (self._key, value))
        return value


class Base(object):
    def __init__(self):
        self.logger = get_ixnet_logger(__name__)

    def create_node(self, ixn_obj, name):
        """It will check/ create a node with name"""
        if name in ixn_obj:
            return ixn_obj.get(name)
        else:
            ixn_obj[name] = list()
            return ixn_obj[name]

    def add_element(self, ixn_obj, name=None):
        ixn_obj.append(dict())
        new_element = ixn_obj[-1]
        new_element["xpath"] = ""
        if name is not None:
            new_element["name"] = self.multivalue(name)
        return new_element

    def create_node_elemet(self, ixn_obj, node_name, name=None):
        """Expectation  of this method:
        - check/ create a node with "node_name"
        - We are setting name as multivalue for farther processing
        - It will return that newly created dict
        """
        self.logger.debug("Creating node for %s" % node_name)
        node = self.create_node(ixn_obj, node_name)
        return self.add_element(node, name)

    def create_property(self, ixn_obj, name):
        ixn_property = ixn_obj[name] = dict()
        ixn_property["xpath"] = ""
        return ixn_property

    def att_dict(self):
        return dict()

    def multivalue(self, value, enum=None):
        if value is not None and enum is not None:
            value = enum[value]
        return MultiValue(value)

    def as_multivalue(self, snappi_obj, name, enum=None):
        return self.multivalue(snappi_obj.get(name), enum)

    def post_calculated(self, key, ref_ixnobj=None, ixnobj=None):
        return PostCalculated(key, ref_ixnobj, ixnobj)

    def get_name(self, object):
        name = object.get("name")
        if isinstance(name, MultiValue):
            name = name.value
        if isinstance(name, list):
            name = name[0]
        return name

    def configure_multivalues(self, snappi_obj, ixn_obj, attr_map):
        """attr_map contains snappi_key : ixn_key/ ixn_info in dict format"""
        self.logger.debug("configuring multivalues:")
        for snappi_attr, ixn_map in attr_map.items():
            if isinstance(ixn_map, dict):
                ixn_attr = ixn_map.get("ixn_attr")
                if ixn_attr is None:
                    raise NameError("ixn_attr is missing within ", ixn_map)
                enum_map = ixn_map.get("enum_map")
                value = snappi_obj.get(snappi_attr)
                if value is None:
                    # We need to specify product default if model not specify
                    default_value = ixn_map.get("default_value")
                    if default_value is None:
                        raise NameError(
                            "Please specify default_value for ", snappi_attr
                        )
                    value = default_value
                if enum_map is not None and value is not None:
                    value = enum_map[value]
                self.logger.debug(
                    "ixn_attr %s with enum value %s" % (ixn_attr, value)
                )
            else:
                ixn_attr = ixn_map
                value = snappi_obj.get(snappi_attr)
                self.logger.debug(
                    "ixn_attr %s with value %s" % (ixn_attr, value)
                )
            ixn_obj[ixn_attr] = self.multivalue(value)

    def configure_multivalues_with_choice(self, snappi_obj, ixn_obj, attr_map):
        """attr_map contains snappi_key : ixn_key/ ixn_info in dict format"""
        self.logger.debug("configuring multivalues:")
        for snappi_attr, ixn_map in attr_map.items():
            ixn_attr = ixn_map
            value = snappi_obj.get(snappi_attr).get("value")
            if snappi_attr == "gateway_mac":
                ixn_obj["resolveGateway"] = self.multivalue(False)
            self.logger.debug("ixn_attr %s with value %s" % (ixn_attr, value))
            ixn_obj[ixn_attr] = self.multivalue(value)

    def get_symmetric_nodes(self, parent_list, node_name):
        nodes_list = []
        max_len = 0
        for parent in parent_list:
            nodes = getattr(parent, node_name)
            node_len = len(nodes)
            if node_len > max_len:
                max_len = node_len
            nodes_list.append(nodes)
        symmetric_nodes = []
        active_list = []
        for nodes in nodes_list:
            if len(nodes) == max_len:
                active_list.extend([True] * max_len)
                symmetric_nodes.extend(nodes)
            else:
                for index in range(0, max_len):
                    node = nodes[0]
                    if index < len(nodes):
                        node = nodes[index]
                        active_list.append(node.active)
                        symmetric_nodes.append(node)
                    else:
                        active_list.append(False)
                        symmetric_nodes.append(node)
        return NodesInfo(max_len, active_list, symmetric_nodes)


class NodesInfo(object):
    def __init__(self, max_len, active_list, symmetric_nodes):
        self._base = Base()
        self._max_len = max_len
        self._active_list = active_list
        self._symmetric_nodes = symmetric_nodes

    @property
    def max_len(self):
        return self._max_len

    @property
    def active_list(self):
        return self._active_list

    @property
    def symmetric_nodes(self):
        return self._symmetric_nodes

    @property
    def is_all_null(self):
        for node in self._symmetric_nodes:
            if node is not None:
                return False
        return True

    def get_values(self, attr_name, enum_map=None, default=None):
        """This will extract values from NodesInfo object.
        It will set default value if value is None and have some default
        Otherwise raise error"""
        values = []
        for node in self._symmetric_nodes:
            value = node.get(attr_name)
            if value is None:
                if default is None:
                    raise NameError(
                        "Please specify default_value for ", attr_name
                    )
                value = default
            if enum_map is not None:
                value = enum_map[value]
            values.append(value)
        return values

    def get_multivalues(self, attr_name, enum_map=None, default=None):
        return self._base.multivalue(
            self.get_values(attr_name, enum_map=enum_map, default=default)
        )

    def config_values(self, ixn_obj, attr_map):
        for snappi_attr, ixn_map in attr_map.items():
            if isinstance(ixn_map, dict):
                ixn_attr = ixn_map.get("ixn_attr")
                if ixn_attr is None:
                    raise NameError("ixn_attr is missing within ", ixn_map)
                enum_map = ixn_map.get("enum_map")
                values = self.get_multivalues(
                    snappi_attr,
                    enum_map=enum_map,
                    default=ixn_map.get("default_value"),
                )
            else:
                ixn_attr = ixn_map
                values = self.get_multivalues(snappi_attr)
            ixn_obj[ixn_attr] = values

    def get_tab(self, tab_name):
        dummy_value = None
        tab_nodes = []
        for node in self._symmetric_nodes:
            tab_node = node.get(tab_name)
            if tab_node is not None:
                dummy_value = tab_node
            tab_nodes.append(tab_node)

        for idx, tab_node in enumerate(tab_nodes):
            if tab_node is None and dummy_value is not None:
                tab_nodes[idx] = dummy_value

        return NodesInfo(self._max_len, self._active_list, tab_nodes)

    def get_symmetric_nodes(self, node_name):
        return self._base.get_symmetric_nodes(self._symmetric_nodes, node_name)

    def get_group_nodes(self, tab_name):
        """We will pass a attribute names which is array in type
        Fill with other nodes and active_list as False
        It will raise error if all elements are not same length
        Finally return list of NodesInfo
        It will use some IxNetwork tab which do not have enable/disable features
        """
        dummy_tab = None
        for node in self._symmetric_nodes:
            dummy_tab = node.get(tab_name)
            if dummy_tab is not None:
                break
        tab_lengths = []
        group_nodes = []
        active_list = []
        for node in self._symmetric_nodes:
            tab = node.get(tab_name)
            if tab is None:
                active_list.append(False)
                if dummy_tab is None:
                    tab_lengths.append(0)
                else:
                    tab_lengths.append(len(dummy_tab))
                    tab = dummy_tab
            else:
                tab_lengths.append(len(tab))
                active_list.append(True)
            if len(set(tab_lengths)) > 1:
                raise Exception(
                    "All the attributes %s should have same lengths" % tab_name
                )
            for idx in range(tab_lengths[-1]):
                if len(group_nodes) <= idx:
                    group_nodes.append([tab[idx]])
                else:
                    group_nodes[idx].append(tab[idx])

        return active_list, [
            NodesInfo(1, self._active_list, group_node)  # use dummy one
            for group_node in group_nodes
        ]

    def get_active_group_nodes(self, tab_name):
        """It will cover get_group_nodes
        At the same time it will raise error if None node
        because IxNetwork do not have active field for those"""
        active_list, node_info_list = self.get_group_nodes(tab_name)
        if len(set(active_list)) > 1:
            raise Exception(
                "All the attributes %s should configure with equal length"
                % tab_name
            )
        return node_info_list



import json
from snappi_ixnetwork.timer import Timer
from snappi_ixnetwork.exceptions import SnappiIxnException


class Lag(object):
    """Transforms OpenAPI objects into IxNetwork objects
    - Lag to /lag
    Args
    ----
    - ixnetworkapi (Api): instance of the Api class
    """

    """
    These are supported keys to defined field (_ETHERNET/ _VLAN/ _LACP/ new one...)
    <stack_name> = {
        <key> : {
            "ixn_attr" : name in IxNetwork SDM
            "default" : default IxNetwork value. Consider snappi default already taken care.
            "translate" : translate snappi value to ixnetwork value help of self method 
            "enum_map" : enum map with snappi with ixnetwork
        }
    }
    """

    _ETHERNET = {
        "mac": {"ixn_attr": "mac", "default": "_default_mac"},
        "mtu": {"ixn_attr": "mtu", "default": "1500"},
    }

    _VLAN_TPID = {
        "x8100": "ethertype8100",
        "x88a8": "ethertype88a8",
        "x9100": "ethertype9100",
        "x9200": "ethertype9200",
        "x9300": "ethertype9300",
    }

    _VLAN = {
        # 'tpid' : {
        #     'ixn_attr' : 'tpid',
        #     'default' : 'ethertype8100',
        #     'enum_map' : _VLAN_TPID
        # },
        "priority": {"ixn_attr": "priority", "default": "0"},
        "vlanId": {"ixn_attr": "vlanId", "default": "1"},
    }

    _LACP = {
        "actor_key": {"ixn_attr": "actorKey", "default": "1"},
        "actor_system_id": {
            "ixn_attr": "actorSystemId",
            "default": "00 00 00 00 00 01",
            "translate": "_translate_actor_system_id",
        },
        "actor_system_priority": {
            "ixn_attr": "actorSystemPriority",
            "default": "1",
        },
    }

    _LACP_PORT_PROTOCOL = {
        "actor_port_number": {"ixn_attr": "actorPortNumber", "default": "1"},
        "actor_port_priority": {
            "ixn_attr": "actorPortPriority",
            "default": "1",
        },
        "lacpdu_periodic_time_interval": {
            "ixn_attr": "lacpduPeriodicTimeInterval",
            "default": "0",
        },
        "lacpdu_timeout": {"ixn_attr": "lacpduTimeout", "default": "0"},
        "actor_activity": {"ixn_attr": "lacpActivity", "default": "active"},
    }

    _STATIC = {"lag_id": {"ixn_attr": "lagId", "default": "1"}}

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self._lag_ports = {}

    def config(self):
        """Transform config.ports into Ixnetwork.Vport
        1) delete any vport that is not part of the config
        2) create a vport for every config.ports[] that is not present in IxNetwork
        3) set config.ports[].location to /vport -location using resourcemanager
        4) set /vport/l1Config/... properties using the corrected /vport -type
        5) connectPorts to use new l1Config settings and clearownership
        """
        self._resource_manager = self._api._ixnetwork.ResourceManager
        self._ixn_lag = self._api._lag
        self._lag_ports = {}
        self._lags_config = self._api.snappi_config.lags
        with Timer(self._api, "Lag Configuration"):
            self._delete_lags()
            if len(self._lags_config) == 0:
                return
            self._create_lags()
        with Timer(self._api, "Lag Ethernet Configuration"):
            self._ethernet_config()
        with Timer(self._api, "Lag Protocol Configuration"):
            self._protocol_config()

    def _import(self, imports):
        if len(imports) > 0:
            errata = self._resource_manager.ImportConfig(
                json.dumps(imports), False
            )
            for item in errata:
                self._api.warning(item)
            return len(errata) == 0
        return True

    def _delete_lags(self):
        """Delete any Lags from the api server that do not exist in the new config"""
        self._api._remove(self._ixn_lag, self._lags_config)

    def _select_lags(self):
        payload = {
            "selects": [
                {
                    "from": "/",
                    "properties": [],
                    "children": [
                        {"child": "lag", "properties": ["name"], "filters": []}
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self._api._ixnetwork.href
        results = self._api._ixnetwork._connection._execute(url, payload)
        lags = {}
        if "lag" in results[0]:
            for lag in results[0]["lag"]:
                lags[lag["name"]] = lag
        return lags

    def _select_vports(self):
        """Select all vports.
        Return them in a dict keyed by vport name.
        """
        payload = {
            "selects": [
                {
                    "from": "/",
                    "properties": [],
                    "children": [
                        {
                            "child": "vport",
                            "properties": [
                                "name",
                            ],
                            "filters": [],
                        }
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self._api._ixnetwork.href
        results = self._api._ixnetwork._connection._execute(url, payload)
        vports = {}
        if "vport" in results[0]:
            for vport in results[0]["vport"]:
                vports[vport["name"]] = vport
        return vports

    def _get_vports(self, ixn_vports, ports):
        vports = []
        for port in ports:
            if port.port_name not in ixn_vports:
                raise Exception("Port {0} is not available".format(port.name))
            vports.append(ixn_vports[port.port_name]["xpath"])
        return vports

    def _create_lags(self):
        """Add any Lags to the api server that do not already exist"""
        self._ixn_lag.find()
        existing_lags = [ixn_lag.Name for ixn_lag in self._ixn_lag]
        for lag in self._lags_config:
            self._lag_ports[lag.name] = lag.ports
            if lag.name not in existing_lags:
                self._ixn_lag.add(Name=lag.name)
        imports = []
        ixn_vports = self._select_vports()
        ixn_lags = self._select_lags()
        for name, ixn_lag in ixn_lags.items():
            self._api.ixn_objects.set(name, ixn_lag)
            lag_import = {
                "xpath": ixn_lag["xpath"],
                "vports": self._get_vports(ixn_vports, self._lag_ports[name]),
            }
            imports.append(lag_import)
        self._import(imports)

    def _set_multivalue(self, source, attribute, value):
        xpath = "/multivalue[@source = '{0} {1}']".format(source, attribute)
        if isinstance(value, list):
            return {"xpath": "{0}/valueList".format(xpath), "values": value}
        else:
            return {"xpath": "{0}/singleValue".format(xpath), "value": value}

    def _ethernet_config(self):
        imports = []
        ixn_lags = self._select_lags()
        for name, ports in self._lag_ports.items():
            ether_xpath = "{0}/protocolStack/ethernet[1]".format(
                ixn_lags[name]["xpath"]
            )
            ixn_lag = self._ixn_lag.find(
                Name="^%s$" % self._api.special_char(name)
            )
            ixn_proto_stack = ixn_lag.ProtocolStack.find()
            if len(ixn_proto_stack) == 0:
                ixn_proto_stack = ixn_lag.ProtocolStack.add()[0]
            ixn_proto_stack.Multiplier = 1
            for eth_attr in Lag._ETHERNET:
                attr_values = self._configure_attribute(
                    eth_attr, Lag._ETHERNET, ports, "ethernet"
                )
                imports.append(
                    self._set_multivalue(
                        ether_xpath,
                        attr_values.ixn_attribute,
                        attr_values.config_value,
                    )
                )
            vlans = self._process_vlans(ports)
            if len(vlans) > 0:
                if len(vlans) > 0:
                    imports.append(
                        self._set_multivalue(ether_xpath, "enableVlans", True)
                    )
                    imports.append(
                        {"xpath": ether_xpath, "vlanCount": len(vlans)}
                    )

                for idx, vlan in enumerate(vlans):
                    for vlan_attr in Lag._VLAN:
                        attr_values = self._configure_attribute(
                            vlan_attr, Lag._VLAN, vlan
                        )
                        vlan_xpath = "{0}/vlan[{1}]".format(
                            ether_xpath, idx + 1
                        )
                        imports.append(
                            self._set_multivalue(
                                vlan_xpath,
                                attr_values.ixn_attribute,
                                attr_values.config_value,
                            )
                        )
        self._import(imports)

    def _lacp_ports_config(self, name, ports):
        imports = []
        ixn_lags = self._select_lags()
        ixn_lag = self._ixn_lag.find(
            Name="^%s$" % self._api.special_char(name)
        )
        ixn_eth = ixn_lag.ProtocolStack.find().Ethernet.find()
        ixn_static = ixn_eth.Lagportstaticlag.find()
        lacp_port_protocols = []
        for port in ports:
            lacp_port_protocol = port.lacp
            lacp_port_protocols.append(lacp_port_protocol)
        if len(ixn_static) > 0:
            ixn_static.remove()
        lacp_xpath = "{0}/protocolStack/ethernet[1]/lagportlacp[1]".format(
            ixn_lags[name]["xpath"]
        )
        imports.append(self._set_multivalue(lacp_xpath, "active", True))
        for lacp_attr in Lag._LACP_PORT_PROTOCOL:
            attr_values = self._configure_attribute(
                lacp_attr, Lag._LACP_PORT_PROTOCOL, lacp_port_protocols
            )
            imports.append(
                self._set_multivalue(
                    lacp_xpath,
                    attr_values.ixn_attribute,
                    attr_values.config_value,
                )
            )
        return imports

    def _protocol_config(self):
        imports = []
        ixn_lags = self._select_lags()
        for snappi_lag in self._lags_config:
            ixn_lag = self._ixn_lag.find(
                Name="^%s$" % self._api.special_char(snappi_lag.name)
            )
            ixn_eth = ixn_lag.ProtocolStack.find().Ethernet.find()
            ixn_lacp = ixn_eth.Lagportlacp.find()
            ixn_static = ixn_eth.Lagportstaticlag.find()
            choice = None
            protocols = []
            protocol = snappi_lag.protocol
            if choice is None:
                choice = protocol.choice
            protocols.append(protocol)
            if choice is None:
                if len(ixn_static) > 0:
                    ixn_static.Active.Single(False)
                if len(ixn_lacp) > 0:
                    ixn_lacp.Active.Single(False)
                return
            if choice == "lacp":
                if len(ixn_static) > 0:
                    ixn_static.remove()
                lacp_xpath = (
                    "{0}/protocolStack/ethernet[1]/lagportlacp[1]".format(
                        ixn_lags[snappi_lag.name]["xpath"]
                    )
                )
                imports.append(
                    self._set_multivalue(lacp_xpath, "active", True)
                )
                for lacp_attr in Lag._LACP:
                    attr_values = self._configure_attribute(
                        lacp_attr, Lag._LACP, protocols, choice
                    )
                    imports.append(
                        self._set_multivalue(
                            lacp_xpath,
                            attr_values.ixn_attribute,
                            attr_values.config_value,
                        )
                    )
                lacp_port_imports = self._lacp_ports_config(
                    snappi_lag.name, snappi_lag.ports
                )
                imports += lacp_port_imports
            else:
                if len(ixn_lacp) > 0:
                    ixn_lacp.remove()
                static_xpath = (
                    "{0}/protocolStack/ethernet[1]/lagportstaticlag[1]".format(
                        ixn_lags[snappi_lag.name]["xpath"]
                    )
                )
                imports.append(
                    self._set_multivalue(static_xpath, "active", True)
                )
                for static_attr in Lag._STATIC:
                    attr_values = self._configure_attribute(
                        static_attr, Lag._STATIC, protocols, choice
                    )
                    imports.append(
                        self._set_multivalue(
                            static_xpath,
                            attr_values.ixn_attribute,
                            attr_values.config_value,
                        )
                    )

        self._import(imports)

    def _process_vlans(self, ports):
        vlan_count = -1
        vlan_list = []
        for port in ports:
            vlans = port.ethernet.vlans
            if vlan_count == -1:
                vlan_count = len(vlans)
                for i in range(len(vlans)):
                    vlan_list.append([vlans[i]])
            elif vlan_count == len(vlans):
                for i in range(len(vlans)):
                    vlan_list[i].append(vlans[i])
            else:
                raise Exception("Please configure equal numbers of VLANs")
        return vlan_list

    def _configure_attribute(self, attr, mapper, parent_list, obj_name=None):
        attr_values = ProtocolAttributes()
        attr_mapper = mapper[attr]
        ixn_attr = attr_mapper["ixn_attr"]
        default_value = attr_mapper["default"]
        enum_map = attr_mapper.get("enum_map")
        translate = attr_mapper.get("translate")
        default_obj = getattr(self, default_value, None)
        attr_values.ixn_attribute = ixn_attr
        for parent in parent_list:
            if obj_name is not None:
                parent = getattr(parent, obj_name, None)
            config_value = getattr(parent, attr, None)
            if config_value is not None:
                if translate is not None:
                    config_value = getattr(self, translate)(config_value)
                if enum_map is None:
                    attr_values.config_value = str(config_value)
                else:
                    attr_values.config_value = enum_map[str(config_value)]
            elif default_obj is None:
                attr_values.config_value = default_value
            else:
                attr_values.config_value = default_obj()
        return attr_values

    def _translate_actor_system_id(self, config_value):
        return config_value.replace(":", " ")

    def _select_protcols(self, lag_href):
        payload = {
            "selects": [
                {
                    "from": lag_href,
                    "properties": [],
                    "children": [
                        {
                            "child": "protocolStack",
                            "properties": ["*"],
                            "filters": [],
                        },
                        {
                            "child": "lagMode",
                            "properties": ["*"],
                            "filters": [],
                        },
                    ],
                    "inlines": [],
                }
            ]
        }
        url = "%s/operations/select?xpath=true" % self._api._ixnetwork.href
        results = self._api._ixnetwork._connection._execute(url, payload)
        lags = {}
        if "lag" in results[0]:
            for lag in results[0]["lag"]:
                lags[lag["name"]] = lag
        return lags

    def _set_result_value(
        self, row, column_name, column_value, column_type=str
    ):
        if (
            len(self._column_names) > 0
            and column_name not in self._column_names
        ):
            return
        try:
            row[column_name] = column_type(column_value)
        except:
            if column_type.__name__ in ["float", "int"]:
                row[column_name] = 0
            else:
                row[column_type] = column_value

    def results(self, request):
        """Return port results"""
        if request.column_names is None:
            self._column_names = []
        else:
            self._column_names = request.column_names
        port_rows = {}
        for vport in self._api.select_vports().values():
            port_row = {}
            self._set_result_value(port_row, "name", vport["name"])
            location = vport["location"]
            if vport["connectionState"].startswith("connectedLink") is True:
                location += ";connected"
            elif len(location) > 0:
                location += ";" + vport["connectionState"]
            else:
                location = vport["connectionState"]
            self._set_result_value(port_row, "location", location)
            self._set_result_value(
                port_row,
                "link",
                (
                    "up"
                    if vport["connectionState"] == "connectedLinkUp"
                    else "down"
                ),
            )
            self._set_result_value(port_row, "capture", "stopped")
            port_rows[vport["name"]] = port_row
        try:
            table = self._api.assistant.StatViewAssistant("Port Statistics")
            for row in table.Rows:
                port_row = port_rows[row["Port Name"]]
                self._set_result_value(
                    port_row, "frames_tx", row["Frames Tx."], int
                )
                self._set_result_value(
                    port_row, "frames_rx", row["Valid Frames Rx."], int
                )
                self._set_result_value(
                    port_row, "frames_tx_rate", row["Frames Tx. Rate"], float
                )
                self._set_result_value(
                    port_row,
                    "frames_rx_rate",
                    row["Valid Frames Rx. Rate"],
                    float,
                )
                self._set_result_value(
                    port_row, "bytes_tx", row["Bytes Tx."], int
                )
                self._set_result_value(
                    port_row, "bytes_rx", row["Bytes Rx."], int
                )
                self._set_result_value(
                    port_row, "bytes_tx_rate", row["Bytes Tx. Rate"], float
                )
                self._set_result_value(
                    port_row, "bytes_rx_rate", row["Bytes Rx. Rate"], float
                )
                self._set_result_value(
                    port_row,
                    "pfc_class_0_frames_rx",
                    row["Rx Pause Priority Group 0 Frames"],
                    int,
                )
                self._set_result_value(
                    port_row,
                    "pfc_class_1_frames_rx",
                    row["Rx Pause Priority Group 1 Frames"],
                    int,
                )
                self._set_result_value(
                    port_row,
                    "pfc_class_2_frames_rx",
                    row["Rx Pause Priority Group 2 Frames"],
                    int,
                )
                self._set_result_value(
                    port_row,
                    "pfc_class_3_frames_rx",
                    row["Rx Pause Priority Group 3 Frames"],
                    int,
                )
                self._set_result_value(
                    port_row,
                    "pfc_class_4_frames_rx",
                    row["Rx Pause Priority Group 4 Frames"],
                    int,
                )
                self._set_result_value(
                    port_row,
                    "pfc_class_5_frames_rx",
                    row["Rx Pause Priority Group 5 Frames"],
                    int,
                )
                self._set_result_value(
                    port_row,
                    "pfc_class_6_frames_rx",
                    row["Rx Pause Priority Group 6 Frames"],
                    int,
                )
                self._set_result_value(
                    port_row,
                    "pfc_class_7_frames_rx",
                    row["Rx Pause Priority Group 7 Frames"],
                    int,
                )
        except:
            pass
        return port_rows.values()


class ProtocolAttributes(object):
    def __init__(self):
        self._ixn_attr = None
        self._config_value = []

    @property
    def ixn_attribute(self):
        return self._ixn_attr

    @ixn_attribute.setter
    def ixn_attribute(self, value):
        self._ixn_attr = value

    @property
    def config_value(self):
        return self._config_value

    @config_value.setter
    def config_value(self, value):
        self._config_value.append(value)



from snappi_ixnetwork.device.base import *
from snappi_ixnetwork.logger import get_ixnet_logger


class CreateIxnConfig(Base):
    def __init__(self, ngpf):
        super(CreateIxnConfig, self).__init__()
        self._ngpf = ngpf
        self.logger = get_ixnet_logger(__name__)
        self._post_calculated_info = list()

    def create(self, node, node_name, parent_xpath=""):
        if not isinstance(node, list):
            raise TypeError("Expecting list to loop through it")
        for idx, element in enumerate(node, start=1):
            if not isinstance(element, dict):
                raise TypeError("Expecting dict")
            xpath = """{parent_xpath}/{node_name}[{index}]""".format(
                parent_xpath=parent_xpath, node_name=node_name, index=idx
            )
            element["xpath"] = xpath
            self._process_element(element, xpath)

    def post_calculate(self):
        for element, key, value in self._post_calculated_info:
            element[key] = value.value

    def _process_element(self, element, parent_xpath, child_name=None):
        if child_name is not None and "xpath" in element:
            child_xpath = """{parent_xpath}/{child_name}""".format(
                parent_xpath=parent_xpath, child_name=child_name
            )
            element["xpath"] = child_xpath
        key_to_remove = []
        for key, value in element.items():
            if key == "name":
                element["name"] = self.get_name(element)
            elif isinstance(value, MultiValue):
                value = self._get_ixn_multivalue(value, key, element["xpath"])
                if value is None:
                    key_to_remove.append(key)
                else:
                    element[key] = value
            elif isinstance(value, PostCalculated):
                self._post_calculated_info.append([element, key, value])
            elif isinstance(value, dict):
                self._process_element(value, parent_xpath, key)
            elif (
                isinstance(value, list)
                and len(value) > 0
                and isinstance(value[0], dict)
            ):
                if child_name is not None:
                    self.create(value, key, element["xpath"])
                else:
                    self.create(value, key, parent_xpath)

        for key in key_to_remove:
            element.pop(key)

    def _get_ixn_multivalue(self, value, att_name, xpath):
        value = value.value
        ixn_value = {
            "xpath": "/multivalue[@source = '{xpath} {att_name}']".format(
                xpath=xpath, att_name=att_name
            )
        }
        if not isinstance(value, list):
            value = [value]
        if len(set(value)) == 1:
            if value[0] is None:
                return None
            else:
                ixn_value["singleValue"] = {
                    "xpath": "/multivalue[@source = '{xpath} {att_name}']/singleValue".format(
                        xpath=xpath, att_name=att_name
                    ),
                    "value": value[0],
                }
                return ixn_value
        else:
            ixn_value["valueList"] = {
                "xpath": "/multivalue[@source = '{xpath} {att_name}']/valueList".format(
                    xpath=xpath, att_name=att_name
                ),
                "values": value,
            }
            return ixn_value



import pytest


@pytest.fixture
def config_v4_devices(api, utils):
    """Configure bgpv4 devices"""
    config = api.config()

    tx, rx = config.ports.port(
        name="tx", location=utils.settings.ports[0]
    ).port(name="rx", location=utils.settings.ports[1])

    tx_device, rx_device = config.devices.device(name="tx_device").device(
        name="rx_device"
    )

    # tx_device config
    tx_eth = tx_device.ethernets.add()
    tx_eth.connection.port_name = tx.name
    tx_eth.name = "tx_eth"
    tx_eth.mac = "00:00:00:00:00:aa"
    tx_ipv4 = tx_eth.ipv4_addresses.add()
    tx_ipv4.name = "tx_ipv4"
    tx_ipv4.address = "21.1.1.2"
    tx_ipv4.prefix = 24
    tx_ipv4.gateway = "21.1.1.1"

    tx_bgpv4 = tx_device.bgp
    tx_bgpv4.router_id = "192.0.0.1"
    tx_bgp4_int = tx_bgpv4.ipv4_interfaces.add()
    tx_bgp4_int.ipv4_name = tx_ipv4.name
    tc_bgp4_peer = tx_bgp4_int.peers.add()
    tc_bgp4_peer.name = "tx_bgpv4"
    tc_bgp4_peer.as_type = "ebgp"
    tc_bgp4_peer.peer_address = "21.1.1.1"
    tc_bgp4_peer.as_number = 65201

    # rx_device config
    rx_eth = rx_device.ethernets.add()
    rx_eth.connection.port_name = rx.name
    rx_eth.name = "rx_eth"
    rx_eth.mac = "00:00:00:00:00:bb"
    rx_ipv4 = rx_eth.ipv4_addresses.add()
    rx_ipv4.name = "rx_ipv4"
    rx_ipv4.address = "21.1.1.1"
    rx_ipv4.prefix = 24
    rx_ipv4.gateway = "21.1.1.2"
    rx_bgpv4 = rx_device.bgp
    rx_bgpv4.router_id = "192.0.0.2"
    rx_bgpv4_int = rx_bgpv4.ipv4_interfaces.add()
    rx_bgpv4_int.ipv4_name = rx_ipv4.name
    rx_bgpv4_peer = rx_bgpv4_int.peers.add()
    rx_bgpv4_peer.name = "rx_bgpv4"
    rx_bgpv4_peer.as_type = "ebgp"
    rx_bgpv4_peer.peer_address = "21.1.1.2"
    rx_bgpv4_peer.as_number = 65200
    rx_rr = rx_bgpv4_peer.v4_routes.add(name="rx_rr")
    rx_rr.addresses.add(count=1000, address="200.1.0.1", prefix=32)

    # flow config
    flow = config.flows.flow(name="convergence_test")[-1]
    flow.tx_rx.device.tx_names = [tx_ipv4.name]
    flow.tx_rx.device.rx_names = [rx_rr.name]

    api.set_config(config)

@pytest.mark.skip(reason="skip to CICD faster")
def test_issue_7(api, config_v4_devices, utils):
    """This unit test to validate the fixes provided for the issue
    https://github.com/open-traffic-generator/snappi-convergence/issues/7"""
    config = api.config()

    tx, rx = config.ports.port(
        name="tx", location=utils.settings.ports[0]
    ).port(name="rx", location=utils.settings.ports[1])

    tx_device, rx_device = config.devices.device(name="tx_device").device(
        name="rx_device"
    )

    # tx_device config
    tx_eth = tx_device.ethernets.add()
    tx_eth.connection.port_name = tx.name
    tx_eth.name = "tx_eth"
    tx_eth.mac = "00:00:00:00:00:aa"
    tx_ipv6 = tx_eth.ipv6_addresses.add()
    tx_ipv6.name = "tx_ipv6"
    tx_ipv6.address = "2000::1"
    tx_ipv6.prefix = 64
    tx_ipv6.gateway = "2000::2"
    tx_bgpv6 = tx_device.bgp
    tx_bgpv6.router_id = "192.0.0.1"
    tx_bgpv6_int = tx_bgpv6.ipv6_interfaces.add()
    tx_bgpv6_int.ipv6_name = tx_ipv6.name
    tx_bgpv6_peer = tx_bgpv6_int.peers.add()
    tx_bgpv6_peer.name = "tx_bgpv6"
    tx_bgpv6_peer.as_type = "ebgp"
    tx_bgpv6_peer.peer_address = "2000::2"
    tx_bgpv6_peer.as_number = 65201

    # rx_device config
    rx_eth = rx_device.ethernets.add()
    rx_eth.connection.port_name = rx.name
    rx_eth.name = "rx_eth"
    rx_eth.mac = "00:00:00:00:00:bb"
    rx_ipv6 = rx_eth.ipv6_addresses.add()
    rx_ipv6.name = "rx_ipv6"
    rx_ipv6.address = "2000::2"
    rx_ipv6.prefix = 64
    rx_ipv6.gateway = "2000::1"
    rx_bgpv6 = rx_device.bgp
    rx_bgpv6.router_id = "192.0.0.2"
    rx_bgpv6_int = rx_bgpv6.ipv6_interfaces.add()
    rx_bgpv6_int.ipv6_name = rx_ipv6.name
    rx_bgpv6_peer = rx_bgpv6_int.peers.add()
    rx_bgpv6_peer.name = "rx_bgpv6"
    rx_bgpv6_peer.as_type = "ebgp"
    rx_bgpv6_peer.peer_address = "2000::1"
    rx_bgpv6_peer.as_number = 65200
    rx6_rr = rx_bgpv6_peer.v6_routes.add(name="rx6_rr")
    rx6_rr.addresses.add(count=1000, address="3000::1", prefix=64)

    # flow config
    flow = config.flows.flow(name="convergence_test")[-1]
    flow.tx_rx.device.tx_names = [tx_ipv6.name]
    flow.tx_rx.device.rx_names = [rx6_rr.name]

    api.set_config(config)

    validate_config(api)


def validate_config(api):
    assert (
        len(api._ixnetwork.Traffic.TrafficItem.find(Name="convergence_test"))
        == 1
    )

    assert (
        len(
            api._ixnetwork.Topology.find()
            .DeviceGroup.find(Name="tx_device")
            .Ethernet.find(Name="tx_eth")
            .Ipv4.find(Name="tx_ipv4")
        )
        == 0
    )

    assert (
        len(
            api._ixnetwork.Topology.find()
            .DeviceGroup.find(Name="tx_device")
            .Ethernet.find(Name="tx_eth")
            .Ipv6.find(Name="tx_ipv6")
        )
        == 1
    )

    assert (
        len(
            api._ixnetwork.Topology.find()
            .DeviceGroup.find(Name="rx_device")
            .Ethernet.find(Name="rx_eth")
            .Ipv4.find(Name="rx_ipv4")
        )
        == 0
    )

    assert (
        len(
            api._ixnetwork.Topology.find()
            .DeviceGroup.find(Name="rx_device")
            .Ethernet.find(Name="rx_eth")
            .Ipv6.find(Name="rx_ipv6")
        )
        == 1
    )


if __name__ == "__main__":
    pytest.main(["-s", __file__])



import pytest


@pytest.mark.skip(
    reason="Fix - intermittent failure as rx ports are not set in capture mode"
)
def test_capture_pcap(serializer, api, tx_port, rx_port, options):
    """Demonstrates how to start capture and get capture results"""
    config = Config(ports=[tx_port, rx_port], options=options)

    # configure capture
    filter = MacAddressFilter(mac="source", filter="000000000000")
    config.captures.append(
        Capture(
            name="rx_capture",
            port_names=[rx_port.name],
            choice=[BasicFilter(filter)],
            enable=True,
        )
    )

    # configure flow
    port_tx_rx = PortTxRx(tx_port_name=tx_port.name, rx_port_name=rx_port.name)
    config.flows.append(
        Flow(
            name="capture",
            tx_rx=TxRx(port_tx_rx),
            packet=[],
            size=Size(128),
            rate=Rate(unit="pps", value=100),
            duration=Duration(FixedPackets(packets=100)),
        )
    )
    api.set_state(State(ConfigState(config=config, state="set")))

    # start capture
    api.set_state(
        State(PortCaptureState(port_names=[rx_port.name], state="start"))
    )

    # start transmit
    api.set_state(State(FlowTransmitState(state="start")))
    time.sleep(5)

    # stop the capture and receive the capture as a stream of bytes
    pcap_bytes = api.get_capture_results(
        CaptureRequest(port_name=rx_port.name)
    )

    # write the pcap bytes to a local file
    with open("%s.pcap" % rx_port.name, "wb") as fid:
        fid.write(pcap_bytes)

    # do stuff using scapy and the pcap file
    from scapy.all import PcapReader

    reader = PcapReader("%s.pcap" % rx_port.name)
    for item in reader:
        print(item.time)
        item.show()


if __name__ == "__main__":
    pytest.main(["-s", __file__])



import re
from collections import namedtuple
from collections.abc import Mapping


def namedtuple_with_defaults(typename, field_names, default_values=()):
    T = namedtuple(typename, field_names)
    T.__new__.__defaults__ = (None,) * len(T._fields)
    if isinstance(default_values, Mapping):
        prototype = T(**default_values)
    else:
        prototype = T(*default_values)
    T.__new__.__defaults__ = tuple(prototype)
    return T


def convert_as_values(as_types, as_values):
    ConvertedAsValues = namedtuple_with_defaults(
        "ConvertedAsValues",
        ("as_num", "as4_num", "ip_addr", "assign_num", "common_num"),
        ([], [], [], [], []),
    )

    convert_values = ConvertedAsValues()
    for idx, as_type in enumerate(as_types):
        num, assign = as_values[idx].split(":")
        convert_values.as_num.append("65101")
        convert_values.as4_num.append("65101")
        convert_values.common_num.append("65101")
        convert_values.ip_addr.append("1.1.1.1")
        convert_values.assign_num.append(assign)
        if as_type == "as":
            convert_values.as_num[idx] = num
            convert_values.common_num[idx] = num
        elif as_type == "as4":
            convert_values.as4_num[idx] = num
            convert_values.common_num[idx] = num
        else:
            convert_values.ip_addr[idx] = num
    return convert_values


def hex_to_ipv4(hex_value):
    bytes = ["".join(x) for x in zip(*[iter(hex_value)] * 2)]
    bytes = [int(x, 16) for x in bytes]
    return ".".join(str(x) for x in reversed(bytes))
